# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Model serializers that operated at file shard level of granularity."""

import abc
import base64
from collections.abc import Callable, Iterable
import concurrent.futures
import itertools
import pathlib
from typing import cast

from typing_extensions import override

from model_signing.hashing import file
from model_signing.hashing import hashing
from model_signing.manifest import manifest
from model_signing.serialization import serialization
from model_signing.serialization import serialize_by_file


def _build_header(*, name: str, start: int, end: int) -> bytes:
    """Builds a header to encode a path with given name and shard range.

    Args:
        entry_name: The name of the entry to build the header for.
        start: Offset for the start of the path shard.
        end: Offset for the end of the path shard.

    Returns:
        A sequence of bytes that encodes all arguments as a sequence of UTF-8
        bytes. Each argument is separated by dots and the last byte is also a
        dot (so the file digest can be appended unambiguously).
    """
    # Prevent confusion if name has a "." inside by encoding to base64.
    encoded_name = base64.b64encode(name.encode("utf-8"))
    encoded_range = f"{start}-{end}".encode("utf-8")
    # Note: empty string at the end, to terminate header with a "."
    return b".".join([encoded_name, encoded_range, b""])


def _endpoints(step: int, end: int) -> Iterable[int]:
    """Yields numbers from `step` to `end` inclusive, spaced by `step`.

    Last value is always equal to `end`, even when `end` is not a multiple of
    `step`. There is always a value returned.

    Examples:
    ```python
    >>> list(_endpoints(2, 8))
    [2, 4, 6, 8]
    >>> list(_endpoints(2, 9))
    [2, 4, 6, 8, 9]
    >>> list(_endpoints(2, 2))
    [2]

    Yields:
        Values in the range, from `step` and up to `end`.
    """
    yield from range(step, end, step)
    yield end


class ShardedFilesSerializer(serialization.Serializer):
    """Generic file shard serializer.

    Traverses the model directory and creates digests for every file found,
    sharding the file in equal shards and computing the digests in parallel.

    Subclasses can then create a manifest with these digests, either listing
    them item by item, combining them into file digests, or combining all of
    them into a single digest.
    """

    def __init__(
        self,
        sharded_hasher_factory: Callable[
            [pathlib.Path, int, int], file.ShardedFileHasher
        ],
        *,
        max_workers: int | None = None,
        allow_symlinks: bool = False,
    ):
        """Initializes an instance to serialize a model with this serializer.

        Args:
            sharded_hasher_factory: A callable to build the hash engine used to
              hash every shard of the files in the model. Because each shard is
              processed in parallel, every thread needs to call the factory to
              start hashing. The arguments are the file, and the endpoints of
              the shard.
            max_workers: Maximum number of workers to use in parallel. Default
              is to defer to the `concurrent.futures` library.
            allow_symlinks: Controls whether symbolic links are included. If a
              symlink is present but the flag is `False` (default) the
              serialization would raise an error.
        """
        self._hasher_factory = sharded_hasher_factory
        self._max_workers = max_workers
        self._allow_symlinks = allow_symlinks

        # Precompute some private values only once by using a mock file hasher.
        # None of the arguments used to build the hasher are used.
        hasher = sharded_hasher_factory(pathlib.Path(), 0, 1)
        self._shard_size = hasher.shard_size

    @override
    def serialize(
        self,
        model_path: pathlib.Path,
        *,
        ignore_paths: Iterable[pathlib.Path] = frozenset(),
    ) -> manifest.Manifest:
        """Serializes the model given by the `model_path` argument.

        Args:
            model_path: The path to the model.
            ignore_paths: The paths to ignore during serialization. If a
              provided path is a directory, all children of the directory are
              ignored.

        Returns:
            The model's serialized `manifest.Manifest`

        Raises:
            ValueError: The model contains a symbolic link, but the serializer
              was not initialized with `allow_symlinks=True`.
        """
        shards = []
        # TODO: github.com/sigstore/model-transparency/issues/200 - When
        # Python3.12 is the minimum supported version, the glob can be replaced
        # with `pathlib.Path.walk` for a clearer interface, and some speed
        # improvement.
        for path in itertools.chain((model_path,), model_path.glob("**/*")):
            serialize_by_file.check_file_or_directory(
                path, allow_symlinks=self._allow_symlinks
            )
            if path.is_file() and not serialize_by_file._ignored(
                path, ignore_paths
            ):
                shards.extend(self._get_shards(path))

        manifest_items = []
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self._max_workers
        ) as tpe:
            futures = [
                tpe.submit(self._compute_hash, model_path, path, start, end)
                for path, start, end in shards
            ]
            for future in concurrent.futures.as_completed(futures):
                manifest_items.append(future.result())

        return self._build_manifest(manifest_items)

    def _get_shards(
        self, path: pathlib.Path
    ) -> list[tuple[pathlib.Path, int, int]]:
        """Determines the shards of a given file path."""
        shards = []
        path_size = path.stat().st_size
        if path_size > 0:
            start = 0
            for end in _endpoints(self._shard_size, path_size):
                shards.append((path, start, end))
                start = end
        return shards

    def _compute_hash(
        self, model_path: pathlib.Path, path: pathlib.Path, start: int, end: int
    ) -> manifest.ShardedFileManifestItem:
        """Produces the manifest item of the file given by `path`.

        Args:
            model_path: The path to the model.
            path: Path to the file in the model, that is currently transformed
              to a manifest item.
            start: The start offset of the shard (included).
            end: The end offset of the shard (not included).

        Returns:
            The itemized manifest.
        """
        relative_path = path.relative_to(model_path)
        digest = self._hasher_factory(path, start, end).compute()
        return manifest.ShardedFileManifestItem(
            path=relative_path, digest=digest, start=start, end=end
        )

    @abc.abstractmethod
    def _build_manifest(
        self, items: Iterable[manifest.ShardedFileManifestItem]
    ) -> manifest.Manifest:
        """Builds an itemized manifest from a given list of items.

        Every subclass needs to implement this method to determine the format of
        the manifest.
        """
        pass


class ManifestSerializer(ShardedFilesSerializer):
    """Model serializers that produces an itemized manifest, at shard level.

    Since the manifest lists each item individually, this will also enable
    support for incremental updates (to be added later).
    """

    @override
    def serialize(
        self,
        model_path: pathlib.Path,
        *,
        ignore_paths: Iterable[pathlib.Path] = frozenset(),
    ) -> manifest.ShardLevelManifest:
        """Serializes the model given by the `model_path` argument.

        The only reason for the override is to change the return type, to be
        more restrictive. This is to signal that the only manifests that can be
        returned are `manifest.ShardLevelManifest` instances.

        Args:
            model_path: The path to the model.
            ignore_paths: The paths to ignore during serialization. If a
              provided path is a directory, all children of the directory are
              ignored.

        Returns:
            The model's serialized `manifest.ShardLevelManifest`

        Raises:
            ValueError: The model contains a symbolic link, but the serializer
              was not initialized with `allow_symlinks=True`.
        """
        return cast(
            manifest.ShardLevelManifest,
            super().serialize(model_path, ignore_paths=ignore_paths),
        )

    @override
    def _build_manifest(
        self, items: Iterable[manifest.ShardedFileManifestItem]
    ) -> manifest.ShardLevelManifest:
        return manifest.ShardLevelManifest(items)


class DigestSerializer(ShardedFilesSerializer):
    """Serializer for a model that performs a traversal of the model directory.

    This serializer produces a single hash for the entire model.
    """

    def __init__(
        self,
        file_hasher_factory: Callable[
            [pathlib.Path, int, int], file.ShardedFileHasher
        ],
        merge_hasher: hashing.StreamingHashEngine,
        *,
        max_workers: int | None = None,
        allow_symlinks: bool = False,
    ):
        """Initializes an instance to serialize a model with this serializer.

        Args:
            hasher_factory: A callable to build the hash engine used to hash
              every shard of the files in the model. Because each shard is
              processed in parallel, every thread needs to call the factory to
              start hashing. The arguments are the file, and the endpoints of
              the shard.
            merge_hasher: A `hashing.StreamingHashEngine` instance used to merge
              individual file shard digests to compute an aggregate digest.
            max_workers: Maximum number of workers to use in parallel. Default
              is to defer to the `concurent.futures` library.
            allow_symlinks: Controls whether symbolic links are included. If a
              symlink is present but the flag is `False` (default) the
              serialization would raise an error.
        """
        super().__init__(
            file_hasher_factory,
            max_workers=max_workers,
            allow_symlinks=allow_symlinks,
        )
        self._merge_hasher = merge_hasher

    @override
    def serialize(
        self,
        model_path: pathlib.Path,
        *,
        ignore_paths: Iterable[pathlib.Path] = frozenset(),
    ) -> manifest.DigestManifest:
        """Serializes the model given by the `model_path` argument.

        The only reason for the override is to change the return type, to be
        more restrictive. This is to signal that the only manifests that can be
        returned are `manifest.DigestManifest` instances.

        Args:
            model_path: The path to the model.
            ignore_paths: The paths to ignore during serialization. If a
              provided path is a directory, all children of the directory are
              ignored.

        Returns:
            The model's serialized `manifest.DigestManifest`

        Raises:
            ValueError: The model contains a symbolic link, but the serializer
              was not initialized with `allow_symlinks=True`.
        """
        return cast(
            manifest.DigestManifest,
            super().serialize(model_path, ignore_paths=ignore_paths),
        )

    @override
    def _build_manifest(
        self, items: Iterable[manifest.ShardedFileManifestItem]
    ) -> manifest.DigestManifest:
        self._merge_hasher.reset()

        for item in sorted(items, key=lambda i: (i.path, i.start, i.end)):
            header = _build_header(
                name=item.path.name, start=item.start, end=item.end
            )
            self._merge_hasher.update(header)
            self._merge_hasher.update(item.digest.digest_value)

        digest = self._merge_hasher.compute()
        return manifest.DigestManifest(digest)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functionality to generate and verify bundles without signing."""

from google.protobuf import json_format
from in_toto_attestation.v1 import statement
from sigstore_protobuf_specs.dev.sigstore.bundle import v1 as bundle_pb
from sigstore_protobuf_specs.dev.sigstore.common import v1 as common_pb
from sigstore_protobuf_specs.io import intoto as intoto_pb

from model_signing.signature.encoding import PAYLOAD_TYPE
from model_signing.signature.signing import Signer
from model_signing.signature.verifying import Verifier


class FakeSigner(Signer):
    """Provides a Signer that just returns the bundle."""

    def sign(self, stmnt: statement.Statement) -> bundle_pb.Bundle:
        env = intoto_pb.Envelope(
            payload=json_format.MessageToJson(stmnt.pb).encode(),
            payload_type=PAYLOAD_TYPE,
            signatures=[intoto_pb.Signature(sig=b"", keyid=None)],
        )
        bdl = bundle_pb.Bundle(
            media_type="application/vnd.dev.sigstore.bundle.v0.3+json",
            verification_material=bundle_pb.VerificationMaterial(
                public_key=common_pb.PublicKey(
                    raw_bytes=None,
                    key_details=common_pb.PublicKeyDetails.PKIX_ECDSA_P256_SHA_256,
                )
            ),
            dsse_envelope=env,
        )
        return bdl


class FakeVerifier(Verifier):
    """Provides a fake verifier that always passes."""

    def verify(self, bundle: bundle_pb.Bundle) -> None:
        pass



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Machinery for representing a serialized representation of an ML model.

When saving a model, or signing an existing one, we build a manifest from the
model (either by using a `serialization.Serializer` or building it from helper
methods that use precomputed values to reduce amount of I/O involved). This is
the manifest used for generating a signature of the model.

When testing the integrity of a model, we start with the model path and model
type. From this, we extract the path to the signed manifest. Next step is to
verify the authenticity of the signature, after which we extract an in-memory
representation of the _expected_ manifest. To verify the integrity of the model,
we run another `serialization.Serializer` instance to compute the _actual_
manifest of the model as we're seeing it. If these two manifests agree then the
model integrity is maintained.

In the simplest case, we are working with `DigestManifest` objects. Here, the
two manifests agree if and only if the digests are the same. A more complex case
is when the manifest itemizes each model component (e.g., every file (or file
shard) is paired with its digest). The manifests agree if every file and digest
matches. Alternatively, we could allow for file renames, and check only the
digests. Optionally, we can have arbitrary logic, saying that the actual
manifest (i.e., files of the model as they are on disk) must match only a subset
of the expected one: e.g., we could have a model saved in multiple formats but
with a single signature. At verification time, only one format is used, so we
only need to verify the corresponding subset. We need to check that we still
have a valid complete model though.

In the future, we envision an API that would allow verification to be done in a
streaming fashion. As the model is loaded for inference, only the integrity of
the part that gets loaded would be verified, to not penalize inference hosts
when only a tiny subset of a large model is used. This also means that we can
add support for updating a manifest in a streaming fashion while a model is
trained (or post-training to update ancillary files -- requires re-signing).

All these alternative scenarios will be implemented by various subclasses of the
`Manifest` class defined here. Itemized manifests make use of `ItemizedManifest`
objects to represent only a part of the model that can be verified individually.
"""

import abc
from collections.abc import Iterable, Iterator
import dataclasses
import pathlib
import sys

from typing_extensions import override

from model_signing.hashing import hashing


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclasses.dataclass(frozen=True)
class ResourceDescriptor:
    """A description of any content from any `Manifest`.

    We aim this to be similar to in-toto's `ResourceDescriptor`. To support
    cases where in-toto cannot be directly used, we make this a dataclass that
    can be mapped to in-toto when needed, and used as its own otherwise.

    Not all fields from in-toto are specified at this moment. All fields here
    must be present, unlike in-toto, where all are optional.

    See github.com/in-toto/attestation/blob/main/spec/v1/resource_descriptor.md
    for the in-toto specification.

    Attributes:
        identifier: A string that uniquely identifies this `ResourceDescriptor`
          within the manifest. Depending on serialized format, users might
          require the identifier to be unique across all manifests stored in a
          system. Producers and consumers can agree on additional requirements
          (e.g., several descriptors must have a common pattern for the
          identifier and the integrity of the model implies integrity of all
          these items, ignoring any other descriptor). Corresponds to `name`,
          `uri`, or `content` in in-toto specification.
        digest: One digest for the item. Note that unlike in-toto, we only have
          one digest for the item and it is always required.
    """

    identifier: str
    digest: hashing.Digest


class Manifest(metaclass=abc.ABCMeta):
    """Generic manifest file to represent a model."""

    @abc.abstractmethod
    def resource_descriptors(self) -> Iterator[ResourceDescriptor]:
        """Yields each resource from the manifest, one by one."""
        pass


@dataclasses.dataclass(frozen=True)
class DigestManifest(Manifest):
    """A manifest that is just a hash."""

    digest: hashing.Digest

    @override
    def resource_descriptors(self) -> Iterator[ResourceDescriptor]:
        """Yields each resource from the manifest, one by one.

        In this case, we have only one descriptor to return. Since model paths
        are already encoded in the digest, use "" for the identifier.
        Subclasses might record additional fields to have distinguishable human
        readable identifiers.
        """
        yield ResourceDescriptor(identifier="", digest=self.digest)


class ItemizedManifest(Manifest):
    """A detailed manifest, recording integrity of every model component."""

    pass


class ManifestItem(metaclass=abc.ABCMeta):
    """An object of a model that can be stored in an `ItemizedManifest`.

    For example, this could be a file, or a file shard. All file paths are
    relative to the model root, to allow moving or renaming the model, without
    invalidating the signature.

    The integrity of each `ManifestItem` can be verified individually and in
    parallel. If the item is backed by a file, we recompute the hash of the
    portion of the file that represents this item.
    """

    pass


@dataclasses.dataclass
class FileManifestItem(ManifestItem):
    """A manifest item that records a filename path together with its digest.

    Note that the path component is a `pathlib.PurePath`, relative to the model.
    """

    path: pathlib.PurePath
    digest: hashing.Digest

    def __init__(self, *, path: pathlib.PurePath, digest: hashing.Digest):
        """Builds a manifest item pairing a file with its digest.

        Args:
            path: The path to the file, relative to the model root.
            digest: The digest of the file.
        """
        # Note: we need to force a PosixPath to canonicalize the manifest.
        self.path = pathlib.PurePosixPath(path)
        self.digest = digest


class FileLevelManifest(ItemizedManifest):
    """A detailed manifest, recording integrity of every model file."""

    def __init__(self, items: Iterable[FileManifestItem]):
        """Builds an itemized manifest from a collection of files.

        Rather than recording the items in a list, we use a dictionary, to allow
        efficient updates and retrieval of digests.
        """
        self._item_to_digest = {item.path: item.digest for item in items}

    def __eq__(self, other: Self):
        return self._item_to_digest == other._item_to_digest

    @override
    def resource_descriptors(self) -> Iterator[ResourceDescriptor]:
        """Yields each resource from the manifest, one by one.

        The items are returned in alphabetical order of the path.
        """
        for item, digest in sorted(self._item_to_digest.items()):
            yield ResourceDescriptor(identifier=str(item), digest=digest)


@dataclasses.dataclass(frozen=True, order=True)
class Shard:
    """A dataclass to hold information about a file shard.

    Attributes:
        path: The path to the file, relative to the model root.
        start: The start offset of the shard (included).
        end: The end offset of the shard (not included).
    """

    path: pathlib.PurePath
    start: int
    end: int

    def __str__(self) -> str:
        """Converts the item to a canonicalized string representation.

        The format is {path}:{start}:{end}, which should also be easy to decode.
        """
        return f"{str(self.path)}:{self.start}:{self.end}"

    @classmethod
    def from_str(cls, s: str) -> Self:
        """Builds a file shard from the string representation.

        It is guaranteed that for a shard `shard` and a (valid) string `s` the
        following two round-trip properties hold:

        ```
        str(Shard.from_str(s)) == s
        Shard.from_str(str(shard)) == shard
        ```

        Raises:
            ValueError: if the string argument does not represent a valid shard
            serialization (is not in the format `path:start:end`).
        """
        parts = s.split(":")
        if len(parts) != 3:
            raise ValueError(f"Expected 3 components separated by `:`, got {s}")

        path = pathlib.PurePosixPath(parts[0])
        start = int(parts[1])
        end = int(parts[2])

        return cls(path, start, end)


@dataclasses.dataclass
class ShardedFileManifestItem(ManifestItem):
    """A manifest item that records a file shard together with its digest."""

    path: pathlib.PurePath
    start: int
    end: int
    digest: hashing.Digest

    def __init__(
        self,
        *,
        path: pathlib.PurePath,
        start: int,
        end: int,
        digest: hashing.Digest,
    ):
        """Builds a manifest item pairing a file shard with its digest.

        Args:
            path: The path to the file, relative to the model root.
            start: The start offset of the shard (included).
            end: The end offset of the shard (not included).
            digest: The digest of the file shard.
        """
        # Note: we need to force a PosixPath to canonicalize the manifest
        self.path = pathlib.PurePosixPath(path)
        self.start = start
        self.end = end
        self.digest = digest

    @property
    def input_tuple(self) -> Shard:
        """Returns the triple that uniquely determines the manifest item."""
        return Shard(self.path, self.start, self.end)


class ShardLevelManifest(FileLevelManifest):
    """A detailed manifest, recording integrity of every model file."""

    def __init__(self, items: Iterable[ShardedFileManifestItem]):
        """Builds an itemized manifest from a collection of files.

        Rather than recording the items in a list, we use a dictionary, to allow
        efficient updates and retrieval of digests.
        """
        self._item_to_digest = {item.input_tuple: item.digest for item in items}

    @override
    def resource_descriptors(self) -> Iterator[ResourceDescriptor]:
        """Yields each resource from the manifest, one by one.

        The items are returned in the order given by the `Shard` dataclass
        (implicit ordering: by file, shard start offset and shard end offset, in
        order).
        """
        for item, digest in sorted(self._item_to_digest.items()):
            yield ResourceDescriptor(identifier=str(item), digest=digest)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from google.protobuf import json_format
from in_toto_attestation.v1 import statement_pb2 as statement_pb


PAYLOAD_TYPE = "application/vnd.in-toto+json"


def pae(
    statement: statement_pb.Statement,  # pylint: disable=no-member
) -> bytes:
    """Generates the PAE encoding of the statement.

    See https://github.com/secure-systems-lab/dsse/blob/v1.0.0/protocol.md
    for details.

    Args:
        statement (statement_pb.Statement): the statement to be encoded.

    Returns:
        bytes: the encoded statement as bytes.
    """
    enc_payload = json_format.MessageToJson(statement).encode()
    payload_len = len(enc_payload)
    pae = (
        "DSSEV1"
        f" {len(PAYLOAD_TYPE)} {PAYLOAD_TYPE}"
        f" {payload_len} {enc_payload}"
    )
    pae = pae.encode()
    return pae



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Support for signing intoto payloads into sigstore bundles."""

import json
import pathlib
import sys

from sigstore_protobuf_specs.dev.sigstore.bundle import v1 as bundle_pb
from typing_extensions import override

from model_signing.manifest import manifest as manifest_module
from model_signing.signature import signing as signature_signing
from model_signing.signature import verifying as signature_verifying
from model_signing.signing import in_toto
from model_signing.signing import signing


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class IntotoSignature(signing.Signature):
    def __init__(self, bundle: bundle_pb.Bundle):
        self._bundle = bundle

    @override
    def write(self, path: pathlib.Path) -> None:
        path.write_text(self._bundle.to_json())

    @classmethod
    @override
    def read(cls, path: pathlib.Path) -> Self:
        bundle = bundle_pb.Bundle().from_json(path.read_text())
        return cls(bundle)

    def to_manifest(self) -> manifest_module.Manifest:
        payload = json.loads(self._bundle.dsse_envelope.payload)
        return in_toto.IntotoPayload.manifest_from_payload(payload)


class IntotoSigner(signing.Signer):
    def __init__(self, sig_signer: signature_signing.Signer):
        self._sig_signer = sig_signer

    @override
    def sign(self, payload: signing.SigningPayload) -> IntotoSignature:
        if not isinstance(payload, in_toto.IntotoPayload):
            raise TypeError("only IntotoPayloads are supported")
        bundle = self._sig_signer.sign(payload.statement)
        return IntotoSignature(bundle)


class IntotoVerifier(signing.Verifier):
    def __init__(self, sig_verifier: signature_verifying.Verifier):
        self._sig_verifier = sig_verifier

    @override
    def verify(self, signature: signing.Signature) -> manifest_module.Manifest:
        if not isinstance(signature, IntotoSignature):
            raise TypeError("only IntotoSignature is supported")
        self._sig_verifier.verify(signature._bundle)
        return signature.to_manifest()



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functionality to sign and verify models with certificates."""

from typing import Self

import certifi
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.x509 import oid as crypto_oid
from in_toto_attestation.v1 import statement
from OpenSSL import crypto as ssl_crypto
from sigstore_protobuf_specs.dev.sigstore.bundle import v1 as bundle_pb
from sigstore_protobuf_specs.dev.sigstore.common import v1 as common_pb

from model_signing.signature.key import ECKeySigner
from model_signing.signature.key import ECKeyVerifier
from model_signing.signature.key import load_ec_private_key
from model_signing.signature.signing import Signer
from model_signing.signature.verifying import VerificationError
from model_signing.signature.verifying import Verifier


def _load_single_cert(path: str) -> x509.Certificate:
    with open(path, "rb") as fd:
        cert = x509.load_pem_x509_certificate(fd.read())
    return cert


def _load_multiple_certs(paths: list[str]) -> list[x509.Certificate]:
    certs = b""
    for p in paths:
        with open(p, "rb") as fd:
            certs += fd.read()
    return x509.load_pem_x509_certificates(certs)


class PKISigner(Signer):
    """Signer using an elliptic curve private key.

    The signer can be used for signing and adds the provided certificate
    information as verification material.
    """

    def __init__(
        self,
        private_key: ec.EllipticCurvePrivateKey,
        signing_cert: x509.Certificate,
        cert_chain: list[x509.Certificate],
    ) -> None:
        self._key_signer = ECKeySigner(private_key)
        self._signing_cert = signing_cert

        pub_key = private_key.public_key()
        cert_pub_key = self._signing_cert.public_key()
        if pub_key != cert_pub_key:
            raise ValueError(
                "the private key's public key does not match the"
                " signing certificates public key"
            )
        self._cert_chain = cert_chain

    @classmethod
    def from_path(
        cls,
        private_key_path: str,
        signing_cert_path: str,
        cert_chain_paths: list[str],
    ) -> Self:
        private_key = load_ec_private_key(private_key_path)
        signing_cert = _load_single_cert(signing_cert_path)
        cert_chain = _load_multiple_certs(cert_chain_paths)
        return cls(private_key, signing_cert, cert_chain)

    @staticmethod
    def __chain(
        signing_cert: x509.Certificate, chain: list[x509.Certificate]
    ) -> list[common_pb.X509Certificate]:
        result_chain = [
            common_pb.X509Certificate(
                raw_bytes=signing_cert.public_bytes(
                    encoding=serialization.Encoding.DER
                )
            )
        ]
        for cert in chain:
            result_chain.append(
                common_pb.X509Certificate(
                    raw_bytes=cert.public_bytes(
                        encoding=serialization.Encoding.DER
                    )
                )
            )
        return result_chain

    def sign(self, stmnt: statement.Statement) -> bundle_pb.Bundle:
        bdl = self._key_signer.sign(stmnt)
        bdl.verification_material.public_key = None
        bdl.verification_material.x509_certificate_chain = (
            common_pb.X509CertificateChain(
                certificates=self.__chain(self._signing_cert, self._cert_chain)
            )
        )
        return bdl


class PKIVerifier(Verifier):
    """Provides a verifier based on root certificates."""

    def __init__(
        self, root_certs: list[x509.Certificate] | None = None
    ) -> None:
        self._store = ssl_crypto.X509Store()
        for c in root_certs:
            self._store.add_cert(ssl_crypto.X509.from_cryptography(c))

    @classmethod
    def from_paths(cls, root_cert_paths: list[str] | None = None) -> Self:
        crypto_trust_roots: list[x509.Certificate] = []
        if root_cert_paths:
            crypto_trust_roots = _load_multiple_certs(root_cert_paths)
        else:
            crypto_trust_roots = _load_multiple_certs([certifi.where()])
        return cls(crypto_trust_roots)

    def verify(self, bundle: bundle_pb.Bundle) -> None:
        signing_chain = bundle.verification_material.x509_certificate_chain
        signing_cert_crypto = x509.load_der_x509_certificate(
            signing_chain.certificates[0].raw_bytes
        )
        sign_time = signing_cert_crypto.not_valid_before_utc
        self._store.set_time(sign_time)
        signing_cert_ossl = ssl_crypto.X509.from_cryptography(
            signing_cert_crypto
        )
        chain = []
        for cert in signing_chain.certificates[1:]:
            chain.append(
                ssl_crypto.X509.from_cryptography(
                    x509.load_der_x509_certificate(cert.raw_bytes)
                )
            )

        store_ctx = ssl_crypto.X509StoreContext(
            self._store, signing_cert_ossl, chain
        )
        try:
            store_ctx.verify_certificate()
        except ssl_crypto.X509StoreContextError as err:
            raise VerificationError(
                f"signing certificate verification failed: {err}"
            ) from err
        usage = signing_cert_crypto.extensions.get_extension_for_class(
            x509.KeyUsage
        )
        if not usage.value.digital_signature:
            raise VerificationError(
                "the certificate is not valid for digital signature usage"
            )
        ext_usage = signing_cert_crypto.extensions.get_extension_for_class(
            x509.ExtendedKeyUsage
        )
        if crypto_oid.ExtendedKeyUsageOID.CODE_SIGNING not in ext_usage.value:
            raise VerificationError(
                "the certificate is not valid for code signing usage"
            )

        # Verify the contents with a key verifier
        pub_key: ec.EllipticCurvePublicKey = signing_cert_crypto.public_key
        verifier = ECKeyVerifier(pub_key)
        return verifier.verify(bundle)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This package provides the functionality to sign models."""

import abc

from in_toto_attestation.v1 import statement
from sigstore_protobuf_specs.dev.sigstore.bundle import v1 as bundle_pb


class Signer(abc.ABC):
    """Signer is the abstract base class for all signing methods."""

    @abc.abstractmethod
    def sign(self, stmnt: statement.Statement) -> bundle_pb.Bundle:
        """Sign signs the provide statment.

        Args:
            stmnt (statement.Statement): The statemnt that needs to be signed.

        Returns:
            bundle_pb.Bundle: DSSE envelop and the verification material.
        """



# Copyright 2024 The Sigstore Authors
# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable, Iterable
import pathlib
from typing import TypeAlias

from model_signing.manifest import manifest
from model_signing.serialization import serialization
from model_signing.signature import verifying
from model_signing.signing import signing


PayloadGeneratorFunc: TypeAlias = Callable[
    [manifest.Manifest], signing.SigningPayload
]


def sign(
    model_path: pathlib.Path,
    signer: signing.Signer,
    payload_generator: PayloadGeneratorFunc,
    serializer: serialization.Serializer,
    ignore_paths: Iterable[pathlib.Path] = frozenset(),
) -> signing.Signature:
    """Provides a wrapper function for the steps necessary to sign a model.

    Args:
        model_path: the model to be signed.
        signer: the signer to be used.
        payload_generator: funtion to generate the manifest.
        serializer: the serializer to be used for the model.
        ignore_paths: paths that should be ignored during serialization.
          Defaults to an empty set.

    Returns:
        The model's signature.
    """
    manifest = serializer.serialize(model_path, ignore_paths=ignore_paths)
    payload = payload_generator(manifest)
    sig = signer.sign(payload)
    return sig


def verify(
    sig: signing.Signature,
    verifier: signing.Verifier,
    model_path: pathlib.Path,
    serializer: serialization.Serializer,
    ignore_paths: Iterable[pathlib.Path] = frozenset(),
):
    """Provides a simple wrapper to verify models.

    Args:
        sig: the signature to be verified.
        verifier: the verifier to verify the signature.
        model_path: the path to the model to compare manifests.
        serializer: the serializer used to generate the local manifest.
        ignore_paths: paths that should be ignored during serialization.
          Defaults to an empty set.

    Raises:
        verifying.VerificationError: on any verification error.
    """
    peer_manifest = verifier.verify(sig)
    local_manifest = serializer.serialize(model_path, ignore_paths=ignore_paths)
    if peer_manifest != local_manifest:
        raise verifying.VerificationError("the manifests do not match")



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sigstore based signature, signers and verifiers."""

import json
import pathlib
import sys

from google.protobuf import json_format
from sigstore import dsse as sigstore_dsse
from sigstore import models as sigstore_models
from sigstore import oidc as sigstore_oidc
from sigstore import sign as sigstore_signer
from sigstore import verify as sigstore_verifier
from typing_extensions import override

from model_signing.hashing import hashing
from model_signing.manifest import manifest
from model_signing.signing import as_bytes
from model_signing.signing import in_toto
from model_signing.signing import signing


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


_IN_TOTO_JSON_PAYLOAD_TYPE: str = "application/vnd.in-toto+json"
_IN_TOTO_STATEMENT_TYPE: str = "https://in-toto.io/Statement/v1"


class SigstoreSignature(signing.Signature):
    """Sigstore signature support, wrapping around `sigstore_models.Bundle`."""

    def __init__(self, bundle: sigstore_models.Bundle):
        """Builds an instance of this signature.

        Args:
            bundle: the Sigstore `Bundle` to wrap around.
        """
        self.bundle = bundle

    @override
    def write(self, path: pathlib.Path) -> None:
        """Writes the signature to disk, to the given path.

        The Sigstore `Bundle` is written in JSON format, per the
        canonicalization defined by the `sigstore-python` library.

        Args:
            path: the path to write the signature to.
        """
        path.write_text(self.bundle.to_json())

    @classmethod
    @override
    def read(cls, path: pathlib.Path) -> Self:
        """Reads the signature from disk.

        Does not perform any signature verification, except what is needed to
        parse the signature file.

        Args:
            path: the path to read the signature from.

        Returns:
            A `SigstoreSignature` object wrapping a Sigstore `Bundle`.

        Raises:
            ValueError: If the Sigstore `Bundle` could not be deserialized from
              the contents of the file pointed to by `path`.
        """
        content = path.read_text()
        return cls(sigstore_models.Bundle.from_json(content))


class SigstoreSigner(signing.Signer):
    """Signing machinery using Sigstore.

    We want to sign both digests and in-toto statements, so we provide two
    separate subclasses for the signing. This class will just handle the common
    parts needed to work with Sigstore.
    """

    def __init__(
        self,
        *,
        oidc_issuer: str | None = None,
        use_ambient_credentials: bool = True,
        use_staging: bool = False,
        identity_token: str | None = None,
    ):
        """Initializes Sigstore signers.

        Needs to set-up a signing context to use the public goods instance and
        machinery for getting an identity token to use in signing.

        Args:
            oidc_issuer: An optional OpenID Connect issuer to use instead of the
              default production one. Only relevant if `use_staging = False`.
              Default is empty, relying on the Sigstore configuration.
            use_ambient_credentials: Use ambient credentials (also known as
              Workload Identity). Default is True. If ambient credentials cannot
              be used (not available, or option disabled), a flow to get signer
              identity via OIDC will start.
            use_staging: Use staging configurations, instead of production. This
              is supposed to be set to True only when testing. Default is False.
            identity_token: An explicit identity token to use when signing,
              taking precedence over any ambient credential or OAuth workflow.
        """
        if use_staging:
            self._signing_context = sigstore_signer.SigningContext.staging()
            self._issuer = sigstore_oidc.Issuer.staging()
        else:
            self._signing_context = sigstore_signer.SigningContext.production()
            if oidc_issuer is not None:
                self._issuer = sigstore_oidc.Issuer(oidc_issuer)
            else:
                self._issuer = sigstore_oidc.Issuer.production()

        self._use_ambient_credentials = use_ambient_credentials
        self._identity_token = identity_token

    def _get_identity_token(self) -> sigstore_oidc.IdentityToken:
        """Obtains an identity token to use in signing.

        The precedence matches that of sigstore-python:
        1) Explicitly supplied identity token
        2) Ambient credential detected in the environment, if enabled
        3) Interactive OAuth flow
        """
        if self._identity_token:
            return sigstore_oidc.IdentityToken(self._identity_token)
        if self._use_ambient_credentials:
            token = sigstore_oidc.detect_credential()
            if token:
                return sigstore_oidc.IdentityToken(token)

        return self._issuer.identity_token(force_oob=True)


class SigstoreArtifactSigner(SigstoreSigner):
    """A Sigstore signer that only signs artifacts.

    In our case, this instance is only used to sign `as_bytes.BytesPayload`
    signing payloads.
    """

    @override
    def sign(self, payload: signing.SigningPayload) -> SigstoreSignature:
        """Signs the provided signing payload.

        Args:
            payload: the payload to sign.

        Returns:
            A `SigstoreSignature` object.

        Raises:
            TypeError: If the `payload` type is not `as_bytes.BytesPayload`.
        """
        if not isinstance(payload, as_bytes.BytesPayload):
            raise TypeError("Only `BytesPayload` payloads are supported")

        token = self._get_identity_token()
        with self._signing_context.signer(token) as signer:
            bundle = signer.sign_artifact(payload.digest)

        return SigstoreSignature(bundle)


class SigstoreDSSESigner(SigstoreSigner):
    """A Sigstore signer that only signs DSSE statements.

    In our case, this instance is only used to sign `in_toto.IntotoPayload`
    signing payloads.
    """

    @override
    def sign(self, payload: signing.SigningPayload) -> SigstoreSignature:
        """Signs the provided signing payload.

        Args:
            payload: the payload to sign.

        Returns:
            A `SigstoreSignature` object.

        Raises:
            TypeError: If the `payload` type is not `as_bytes.BytesPayload`.
        """
        if not isinstance(payload, in_toto.IntotoPayload):
            raise TypeError("Only `IntotoPayload` payloads are supported")

        # We need to convert from in-toto statement to Sigstore's DSSE
        # version. They both contain the same contents, but there is no way
        # to coerce one type to the other.
        # See also: https://github.com/sigstore/sigstore-python/issues/1076
        statement = sigstore_dsse.Statement(
            json_format.MessageToJson(payload.statement.pb).encode("utf-8")
        )

        token = self._get_identity_token()
        with self._signing_context.signer(token) as signer:
            bundle = signer.sign_dsse(statement)

        return SigstoreSignature(bundle)


class SigstoreVerifier(signing.Verifier):
    """Signature verification machinery using Sigstore.

    We want to be able to verify signatures generated from either digests or
    in-toto statements, so we provide two separate subclasses for the
    verification. This class will just handle the common parts needed to work
    with Sigstore.
    """

    def __init__(
        self, *, identity: str, oidc_issuer: str, use_staging: bool = False
    ):
        """Initializes Sigstore verifiers.

        When verifying a signature, we also check an identity policy: the
        certificate must belong to a given "identity", and must be issued by a
        given OpenID Connect issuer.

        Args:
            identity: The expected identity that has signed the model.
            oidc_issuer: The expected OpenID Connect issuer that provided the
              certificate used for the signature.
            use_staging: Use staging configurations, instead of production. This
              is supposed to be set to True only when testing. Default is False.
        """
        if use_staging:
            self._verifier = sigstore_verifier.Verifier.staging()
        else:
            self._verifier = sigstore_verifier.Verifier.production()

        # TODO: https://github.com/sigstore/model-transparency/issues/271 -
        # Support additional verification policies
        self._policy = sigstore_verifier.policy.Identity(
            identity=identity, issuer=oidc_issuer
        )


class SigstoreArtifactVerifier(SigstoreVerifier):
    """A Sigstore verifier for signatures on simple digests.

    This class only accepts signatures generated by `SigstoreArtifactSigner`.
    """

    def __init__(
        self,
        expected_digest: bytes,
        *,
        identity: str,
        oidc_issuer: str,
        use_staging: bool = False,
    ):
        """Initializes this verifier.

        Since the signature was generated over a digest, we need to pass the
        expected digest as an argument here (as that is what Sigstore's
        `verify_artifact` expects). Hence, in order to use this class, the model
        on disk needs to be first serialized and the obtained digest can be
        passed to signature verification. This is in reverse compared to
        manifest based signatures where signature verification results in a
        manifest of all expected files and then other layers in the library can
        verify the model integrity (optionally with additional policies).

        Args:
            expected_digest: Expected digest. Must match what was signed.
            identity: The expected identity that has signed the model.
            oidc_issuer: The expected OpenID Connect issuer that provided the
              certificate used for the signature.
            use_staging: Use staging configurations, instead of production. This
              is supposed to be set to True only when testing. Default is False.
        """
        super().__init__(
            identity=identity, oidc_issuer=oidc_issuer, use_staging=use_staging
        )
        self._expected_digest = expected_digest

    @override
    def verify(self, signature: signing.Signature) -> manifest.DigestManifest:
        """Verifies the signature.

        Args:
            signature: the signature to verify.

        Returns:
            A `manifest.DigestManifest` instance that represents the model as a
            single hash. If the function returns without raising an exception,
            then verification succeeded.

        Raises:
            ValueError: If the signature verification fails.
            TypeError: If the signature type is not `SigstoreSignature`.
        """
        if not isinstance(signature, SigstoreSignature):
            raise TypeError("Only `SigstoreSignature` signatures are supported")

        self._verifier.verify_artifact(
            input_=self._expected_digest,
            bundle=signature.bundle,
            policy=self._policy,
        )

        digest = hashing.Digest("sha256", self._expected_digest)
        return manifest.DigestManifest(digest)


class SigstoreDSSEVerifier(SigstoreVerifier):
    """A Sigstore verifier for signatures on simple digests.

    This class only accepts signatures generated by `SigstoreArtifactSigner`.
    """

    @override
    def verify(self, signature: signing.Signature) -> manifest.Manifest:
        """Verifies the signature.

        Args:
            signature: the signature to verify.

        Returns:
            A manifest that represents the model when it was signed.

        Raises:
            ValueError: If the signature verification fails, or if the DSSE
              envelope is not in the expected format.
            TypeError: If the signature type is not `SigstoreSignature`.
        """
        if not isinstance(signature, SigstoreSignature):
            raise TypeError("Only `SigstoreSignature` signatures are supported")

        payload_type, payload = self._verifier.verify_dsse(
            bundle=signature.bundle, policy=self._policy
        )

        if payload_type != _IN_TOTO_JSON_PAYLOAD_TYPE:
            raise ValueError(
                f"Expected DSSE payload {_IN_TOTO_JSON_PAYLOAD_TYPE}, "
                f"but got {payload_type}"
            )

        payload = json.loads(payload)

        if payload["_type"] != _IN_TOTO_STATEMENT_TYPE:
            raise ValueError(
                f"Expected in-toto {_IN_TOTO_STATEMENT_TYPE} payload, "
                f"but got {payload['_type']}"
            )

        return in_toto.IntotoPayload.manifest_from_payload(payload)



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Public, high-level API for the model_signing library.

Users should use this API to sign models and verify the model integrity instead
of reaching out to the internals of the library. We guarantee backwards
compatibility only for the API defined in this file.
"""

from collections.abc import Callable, Iterable
import os
import pathlib
import sys
from typing import Literal, cast

from model_signing.hashing import file
from model_signing.hashing import hashing
from model_signing.hashing import memory
from model_signing.manifest import manifest
from model_signing.serialization import serialize_by_file
from model_signing.serialization import serialize_by_file_shard
from model_signing.signing import in_toto
from model_signing.signing import signing
from model_signing.signing import sigstore


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


def hash(model_path: os.PathLike) -> manifest.Manifest:
    """Hashes a model using the default configuration.

    We use a separate method and configuration for hashing as it needs to be
    common between signing and signature verification. Having thise separate
    also helps with performance testing, as hashing is expected to take the
    largest amount of time (proportional to model size).

    Since we need to be flexible on the serialization format, this returns a
    manifest, instead of just a single digest. The type of returned manifest
    depends on the configuration.

    Args:
        model_path: the path to the model to hash.

    Returns:
        A manifest of the hashed model.
    """
    return HashingConfig().hash(model_path)


def sign(model_path: os.PathLike, signature_path: os.PathLike):
    """Signs a model using the default configuration.

    Args:
        model_path: the path to the model to sign.
        signature_path: the path of the resulting signature.
    """
    SigningConfig().sign(model_path, signature_path)


def verify(
    model_path: os.PathLike,
    signature_path: os.PathLike,
    *,
    identity: str,
    oidc_issuer: str | None = None,
    use_staging: bool = False,
):
    """Verifies that a model conforms to a signature.

    Currently, this assumes signatures over DSSE, using Sigstore. We will add
    support for more cases in a future change.

    Args:
        model_path: the path to the model to verify.
        signature_path: the path to the signature to check.
        identity: The expected identity that has signed the model.
        oidc_issuer: The expected OpenID Connect issuer that provided the
          certificate used for the signature.
        use_staging: Use staging configurations, instead of production. This
          is supposed to be set to True only when testing. Default is False.
    """
    VerificationConfig().set_sigstore_dsse_verifier(
        identity=identity, oidc_issuer=oidc_issuer, use_staging=use_staging
    ).verify(model_path, signature_path)


class HashingConfig:
    """Configuration to use when hashing models.

    Hashing a model results in a `manifest.Manifest` object. This may contain a
    single digest for the entire model, or be a pairing between model components
    (e.g., files, file shards, etc.) and their corresponding hash.

    This configuration class allows selecting the serialization method to
    generate the desired manifest format.

    This configuration class also allows configuring files from within the model
    directory that should be ignored. These are files that doesn't impact the
    behavior of the model, or files that won't be distributed with the model.

    Note that currently this configuration class only supports the main options
    provided by the library. For more granular choices, usage of the lower level
    APIs is recommended.
    """

    def __init__(self):
        """Initializes the default configuration for hashing.

        The default hashing configuration uses SHA256 to compute the digest of
        every file in the model. The resulting manifest is a listing of files
        paired with their hashes. By default, no file is ignored and any
        symbolic link in the model directory results in an error.
        """
        self._ignored_paths = frozenset()
        self._serializer = serialize_by_file.ManifestSerializer(
            self._build_file_hasher_factory(), allow_symlinks=False
        )

    def hash(self, model_path: os.PathLike) -> manifest.Manifest:
        """Hashes a model using the current configuration."""
        return self._serializer.serialize(
            pathlib.Path(model_path), ignore_paths=self._ignored_paths
        )

    def _build_stream_hasher(
        self, hashing_algorithm: Literal["sha256", "blake2"] = "sha256"
    ) -> hashing.StreamingHashEngine:
        """Builds a streaming hasher from a constant string.

        Args:
            hashing_algorithm: the hashing algorithm to use.

        Returns:
            An instance of the requested hasher.
        """
        match hashing_algorithm:
            case "sha256":
                return memory.SHA256()
            case "blake2":
                return memory.BLAKE2()
            case _:
                raise ValueError(
                    f"Unsupported hashing method {hashing_algorithm}"
                )

    def _build_file_hasher_factory(
        self,
        hashing_algorithm: Literal["sha256", "blake2"] = "sha256",
        chunk_size: int = 8192,
    ) -> Callable[[pathlib.Path], file.SimpleFileHasher]:
        """Builds the hasher factory for a serialization by file.

        Args:
            hashing_algorithm: the hashing algorithm to use to hash a file
            chunk_size: The amount of file to read at once. Default is 8KB. A
              special value of 0 signals to attempt to read everything in a
              single call.

        Returns:
            The hasher factory that should be used by the active serialization
            method.
        """

        def factory(path: pathlib.Path) -> file.SimpleFileHasher:
            hasher = self._build_stream_hasher(hashing_algorithm)
            return file.SimpleFileHasher(path, hasher, chunk_size=chunk_size)

        return factory

    def _build_sharded_file_hasher_factory(
        self,
        hashing_algorithm: Literal["sha256", "blake2"] = "sha256",
        chunk_size: int = 8192,
        shard_size: int = 1000000,
    ) -> Callable[[pathlib.Path, int, int], file.ShardedFileHasher]:
        """Builds the hasher factory for a serialization by file shards.

        Args:
            hashing_algorithm: the hashing algorithm to use to hash a file
            chunk_size: The amount of file to read at once. Default is 8KB. A
              special value of 0 signals to attempt to read everything in a
              single call.
            shard_size: The size of a file shard. Default is 1,000,000 bytes.

        Returns:
            The hasher factory that should be used by the active serialization
            method.
        """
        algorithm = self._build_stream_hasher(hashing_algorithm)

        def factory(
            path: pathlib.Path, start: int, end: int
        ) -> file.ShardedFileHasher:
            return file.ShardedFileHasher(
                path,
                algorithm,
                start=start,
                end=end,
                chunk_size=chunk_size,
                shard_size=shard_size,
            )

        return factory

    def set_serialize_by_file_to_manifest(
        self,
        *,
        hashing_algorithm: Literal["sha256", "blake2"] = "sha256",
        chunk_size: int = 8192,
        max_workers: int | None = None,
        allow_symlinks: bool = False,
    ) -> Self:
        """Configures serialization to a manifest pairing files with hashes.

        The serialization method in this configuration is changed to one where
        every file in the model is paired with its digest and a manifest
        containing all these pairings is being returned.

        Args:
            hashing_algorithm: the hashing algorithm to use to hash a file
            chunk_size: The amount of file to read at once. Default is 8KB. A
              special value of 0 signals to attempt to read everything in a
              single call.
            max_workers: Maximum number of workers to use in parallel. Default
              is to defer to the `concurrent.futures` library.
            allow_symlinks: Controls whether symbolic links are included. If a
              symlink is present but the flag is `False` (default) the
              serialization would raise an error.

        Returns:
            The new hashing configuration with the new serialization method.
        """
        self._serializer = serialize_by_file.ManifestSerializer(
            self._build_file_hasher_factory(hashing_algorithm, chunk_size),
            max_workers=max_workers,
            allow_symlinks=allow_symlinks,
        )
        return self

    def set_serialize_by_file_to_digest(
        self,
        *,
        hashing_algorithm: Literal["sha256", "blake2"] = "sha256",
        merge_algorithm: Literal["sha256", "blake2"] = "sha256",
        chunk_size: int = 8192,
        allow_symlinks: bool = False,
    ) -> Self:
        """Configures serialization to a single digest, at file granularity.

        The serialization method in this configuration is changed to one where
        every file in the model is paired with its digest and then a single
        digest is computed over this pairing.

        Args:
            hashing_algorithm: the hashing algorithm to use to hash a file
            merge_algorithm: the hashing algorithm to use when computing the
              final digest over all the (file, digest) pairings
            chunk_size: The amount of file to read at once. Default is 8KB. A
              special value of 0 signals to attempt to read everything in a
              single call.
            allow_symlinks: Controls whether symbolic links are included. If a
              symlink is present but the flag is `False` (default) the
              serialization would raise an error.

        Returns:
            The new hashing configuration with the new serialization method.
        """
        # TODO: https://github.com/sigstore/model-transparency/issues/197 -
        # Because the API for this case is different than the other ones, we
        # have to perform additional steps here.
        file_hasher = cast(
            file.SimpleFileHasher,
            self._build_file_hasher_factory(
                hashing_algorithm, chunk_size=chunk_size
            )(pathlib.Path("unused")),
        )
        merge_hasher = self._build_stream_hasher(merge_algorithm).__class__
        self._serializer = serialize_by_file.DigestSerializer(
            file_hasher, merge_hasher, allow_symlinks=allow_symlinks
        )
        return self

    def set_serialize_by_file_shard_to_manifest(
        self,
        *,
        hashing_algorithm: Literal["sha256", "blake2"] = "sha256",
        chunk_size: int = 8192,
        shard_size: int = 1000000,
        max_workers: int | None = None,
        allow_symlinks: bool = False,
    ) -> Self:
        """Configures serialization to a manifest of (file shard, hash) pairs.

        The serialization method in this configuration is changed to one where
        every file in the model is sharded in equal sized shards and every shard
        is paired with its digest and a manifest containing all these pairings
        is being returned.

        Args:
            hashing_algorithm: the hashing algorithm to use to hash a file shard
            chunk_size: The amount of file to read at once. Default is 8KB. A
              special value of 0 signals to attempt to read everything in a
              single call.
            shard_size: The size of a file shard. Default is 1,000,000 bytes.
            max_workers: Maximum number of workers to use in parallel. Default
              is to defer to the `concurrent.futures` library.
            allow_symlinks: Controls whether symbolic links are included. If a
              symlink is present but the flag is `False` (default) the
              serialization would raise an error.

        Returns:
            The new hashing configuration with the new serialization method.
        """
        self._serializer = serialize_by_file_shard.ManifestSerializer(
            self._build_sharded_file_hasher_factory(
                hashing_algorithm, chunk_size, shard_size
            ),
            max_workers=max_workers,
            allow_symlinks=allow_symlinks,
        )
        return self

    def set_serialize_by_file_shard_to_digest(
        self,
        *,
        hashing_algorithm: Literal["sha256", "blake2"] = "sha256",
        merge_algorithm: Literal["sha256", "blake2"] = "sha256",
        chunk_size: int = 8192,
        shard_size: int = 1000000,
        max_workers: int | None = None,
        allow_symlinks: bool = False,
    ) -> Self:
        """Configures serialization to a single digest, at shard granularity.

        The serialization method in this configuration is changed to one where
        every file shard in the model is paired with its digest and then a
        single digest is computed over all entries in this pairing.

        Args:
            hashing_algorithm: the hashing algorithm to use to hash a file shard
            merge_algorithm: the hashing algorithm to use when computing the
              final digest over all the (file, digest) pairings
            chunk_size: The amount of file to read at once. Default is 8KB. A
              special value of 0 signals to attempt to read everything in a
              single call.
            shard_size: The size of a file shard. Default is 1,000,000 bytes.
            max_workers: Maximum number of workers to use in parallel. Default
              is to defer to the `concurrent.futures` library.
            allow_symlinks: Controls whether symbolic links are included. If a
              symlink is present but the flag is `False` (default) the
              serialization would raise an error.

        Returns:
            The new hashing configuration with the new serialization method.
        """
        merge_hasher = self._build_stream_hasher(merge_algorithm)
        self._serializer = serialize_by_file_shard.DigestSerializer(
            self._build_sharded_file_hasher_factory(
                hashing_algorithm, chunk_size, shard_size
            ),
            merge_hasher,
            max_workers=max_workers,
            allow_symlinks=allow_symlinks,
        )
        return self

    def set_ignored_paths(self, paths: Iterable[os.PathLike]) -> Self:
        """Configures the paths to be ignored during serialization of a model.

        If the model is a single file, there are no paths that are ignored. If
        the model is a directory, all paths must be within the model directory.
        If a path to be ignored is absolute, we convert it to a path within the
        model directory during serialization. If the path is relative, it is
        assumed to be relative to the model root.

        If a path is a directory, serialization will ignore both the path and
        any of its children.

        Args:
            paths: the paths to ignore

        Returns:
            The new hashing configuration with a new set of ignored paths.
        """
        self._ignored_paths = frozenset({pathlib.Path(p) for p in paths})
        return self


class SigningConfig:
    """Configuration to use when signing models.

    The signing configuration is used to decouple between serialization formats
    and signing types. This configuration class allows setting up the
    serialization format, the method to convert a `manifest.Manifest` to a
    signing payload and the engine used for signing (currently, only supporting
    Sigstore at this level).
    """

    def __init__(self):
        """Initializes the default configuration for signing."""
        self._hashing_config = HashingConfig()
        self._payload_generator = in_toto.DigestsIntotoPayload.from_manifest
        self._signer = sigstore.SigstoreDSSESigner(
            use_ambient_credentials=False, use_staging=False
        )

    def sign(self, model_path: os.PathLike, signature_path: os.PathLike):
        """Signs a model using the current configuration.

        Args:
            model_path: the path to the model to sign.
            signature_path: the path of the resulting signature.
        """
        manifest = self._hashing_config.hash(model_path)
        payload = self._payload_generator(manifest)
        signature = self._signer.sign(payload)
        signature.write(pathlib.Path(signature_path))

    def set_hashing_config(self, hashing_config: HashingConfig) -> Self:
        """Sets the new configuration for hashing models.

        Args:
            hashing_config: the new hashing configuration.

        Returns:
            The new signing configuration.
        """
        self._hashing_config = hashing_config
        return self

    def set_payload_generator(
        self, generator: Callable[[manifest.Manifest], signing.SigningPayload]
    ) -> Self:
        """Sets the conversion from manifest to signing payload.

        Since we want to support multiple serialization formats and multiple
        signing solutions, we use a payload generator to relax the coupling
        between the two.

        Args:
            generator: the conversion function from a `manifest.Manifest` to a
              `signing.SigningPayload` payload.

        Return:
            The new signing configuration.
        """
        self._payload_generator = generator
        return self

    def set_sigstore_signer(
        self,
        *,
        sign_dsse: bool = True,
        oidc_issuer: str | None = None,
        use_ambient_credentials: bool = True,
        use_staging: bool = False,
        identity_token: str | None = None,
    ) -> Self:
        """Configures the signing to be performed with Sigstore.

        Only one signer can be configured. Currently, we only support Sigstore
        in the API, but the CLI supports signing with PKI, BYOK and no signing.
        We will merge the configurations in a subsequent change.

        Args:
            sign_dsse: Sign a DSSE statement (if True) or a binary blob.
            oidc_issuer: An optional OpenID Connect issuer to use instead of the
              default production one. Only relevant if `use_staging = False`.
              Default is empty, relying on the Sigstore configuration.
            use_ambient_credentials: Use ambient credentials (also known as
              Workload Identity). Default is True. If ambient credentials cannot
              be used (not available, or option disabled), a flow to get signer
              identity via OIDC will start.
            use_staging: Use staging configurations, instead of production. This
              is supposed to be set to True only when testing. Default is False.
            identity_token: An explicit identity token to use when signing,
              taking precedence over any ambient credential or OAuth workflow.

        Return:
            The new signing configuration.
        """
        if sign_dsse:
            signer_factory = sigstore.SigstoreDSSESigner
        else:
            signer_factory = sigstore.SigstoreArtifactSigner

        self._signer = signer_factory(
            oidc_issuer=oidc_issuer,
            use_ambient_credentials=use_ambient_credentials,
            use_staging=use_staging,
            identity_token=identity_token,
        )
        return self


class VerificationConfig:
    """Configuration to use when verifying models against signatures.

    The signing configuration is used to decouple between serialization formats
    and signing types. This configuration class allows setting up the
    serialization format, the method to convert a `manifest.Manifest` to a
    signing payload and the engine used for signing (currently, only supporting
    Sigstore at this level).
    """

    def __init__(self):
        """Initializes the default configuration for verification."""
        self._hashing_config = HashingConfig()
        self._verifier = None

    def verify(self, model_path: os.PathLike, signature_path: os.PathLike):
        """Verifies that a model conforms to a signature.

        Args:
            model_path: the path to the model to verify.
            signature_path: the path to the signature to check.
        """
        signature = sigstore.SigstoreSignature.read(
            pathlib.Path(signature_path)
        )
        expected_manifest = self._verifier.verify(signature)
        actual_manifest = self._hashing_config.hash(model_path)

        if actual_manifest != expected_manifest:
            raise ValueError("Signature mismatch")

    def set_hashing_config(self, hashing_config: HashingConfig) -> Self:
        """Sets the new configuration for hashing models.

        Args:
            hashing_config: the new hashing configuration.

        Returns:
            The new signing configuration.
        """
        self._hashing_config = hashing_config
        return self

    def set_sigstore_dsse_verifier(
        self,
        *,
        identity: str,
        oidc_issuer: str | None = None,
        use_staging: bool = False,
    ) -> Self:
        """Configures the verification of a Sigstore signature over DSSE.

        Only one verifier can be configured. Currently, we only support Sigstore
        in the API, but the CLI supports signing with PKI, BYOK and no
        signing/verification.  We will merge the configurations in a subsequent
        change.

        Args:
            identity: The expected identity that has signed the model.
            oidc_issuer: The expected OpenID Connect issuer that provided the
              certificate used for the signature.
            use_staging: Use staging configurations, instead of production. This
              is supposed to be set to True only when testing. Default is False.

        Return:
            The new verification configuration.
        """
        self._verifier = sigstore.SigstoreDSSEVerifier(
            identity=identity, oidc_issuer=oidc_issuer, use_staging=use_staging
        )
        return self



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Machinery for signing and verification of ML models.

The serialization API produces a manifest representation of the models, and we
use that to implement integrity checking of models in different computational
patterns. This means that all manifests need to be kept only in memory.

Hence, for signing, we need a separate class hierarchy to represent the payload.
This is why we introduce `SigningPayload` abstract class here. Every instance of
this class is built via a `from_manifest` method (to allow for classes that need
initialization to be performed only once in their constructor but then convert
multiple manifests into payloads).

Since we need to support multiple signing methods (e.g., Sigstore, own PKI,
BCID), we provide a `Signer` abstract class with a single `sign` method that
takes a signing payload and converts it to a signature in the supported format.
Since signers may only accept payloads in specific formats, the `sign` method
can raise a `TypeError` if the provided `SigningPayload` instance is not
supported (due to typing rules, we cannot just use a subclass type for the
argument).

Every possible signature will be implemented as a subclass of `Signature` class.
The API for signatures only allows writing them to disk and parsing them from a
given path.

Finally, every signature needs to be verified. We pair every `Signer` subclass
with a `Verifier` which takes a signature, verify the authenticity of the
payload and then expand that to a `manifest.Manifest` subclass.
"""

import abc
import pathlib
import sys

from model_signing.manifest import manifest


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class SigningPayload(metaclass=abc.ABCMeta):
    """Generic payload that we can sign."""

    @classmethod
    @abc.abstractmethod
    def from_manifest(cls, manifest: manifest.Manifest) -> Self:
        """Converts a manifest to the signing payload used for signing.

        Args:
            manifest: the manifest to convert to signing payload.

        Returns:
            An instance of `SigningPayload` or subclass of it.
        """
        pass


class Signature(metaclass=abc.ABCMeta):
    """Generic signature support."""

    @abc.abstractmethod
    def write(self, path: pathlib.Path) -> None:
        """Writes the signature to disk, to the given path.

        Args:
            path: the path to write the signature to.
        """
        pass

    @classmethod
    @abc.abstractmethod
    def read(cls, path: pathlib.Path) -> Self:
        """Reads the signature from disk.

        Does not perform any signature verification, except what is needed to
        parse the signature file.

        Args:
            path: the path to read the signature from.

        Returns:
            An instance of the class which can be passed to a `Verifier` for
            signature and integrity verification.

        Raises:
            ValueError: If the provided path is not deserializable to the format
              expected by the `Signature` (sub)class.
        """
        pass


class Signer(metaclass=abc.ABCMeta):
    """Generic signer for `SigningPayload` objects.

    Every signer is allowed to only support some signing payload formats. Every
    signer produces a signature in its own format. No signer is required to
    support all subclasses of `SigningPayload` or `Signature`.

    Each signer may implement its own mechanism for managing the key material.
    """

    @abc.abstractmethod
    def sign(self, payload: SigningPayload) -> Signature:
        """Signs the provided signing payload.

        Args:
            payload: the `SigningPayload` instance that should be signed.

        Returns:
            A valid signature.

        Raises:
            TypeError: If the `payload` type is not one of the subclasses of
              `SigningPayload` that are supported.
        """
        pass


class Verifier(metaclass=abc.ABCMeta):
    """Generic signature verifier.

    Every subclass of `Verifier` is paired with a subclass of `Signer`. This is
    to ensure that they support the same signing payload and signature formats
    as well as have similar key materials.

    If the signature is valid, the payload is expanded to a `Manifest` instance
    which can then be used to check the model integrity.
    """

    @abc.abstractmethod
    def verify(self, signature: Signature) -> manifest.Manifest:
        """Verifies the signature.

        Args:
            signature: the signature to verify.

        Returns:
            A `manifest.Manifest` instance that represents the model.

        Raises:
            ValueError: If the signature verification fails.
            TypeError: If the signature is not one of the `Signature` subclasses
              accepted by the verifier.
        """
        pass



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Machinery for serializing ML models.

Currently we have only one serializer that performs a DFS traversal of the model
directory, but more serializers are coming soon.
"""

import abc
from collections.abc import Iterable
import pathlib

from model_signing.manifest import manifest


class Serializer(metaclass=abc.ABCMeta):
    """Generic ML model format serializer."""

    @abc.abstractmethod
    def serialize(
        self,
        model_path: pathlib.Path,
        *,
        ignore_paths: Iterable[pathlib.Path] = frozenset(),
    ) -> manifest.Manifest:
        """Serializes the model given by the `model_path` argument.

        Args:
            model_path: The path to the model.
            ignore_paths: The paths to ignore during serialization. If a
              provided path is a directory, all children of the directory are
              ignored.

        Returns:
            The model's serialized `manifest.Manifest`
        """
        pass



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Signing payload that is just a bytes view on the digest.

In general, this should only be used if we want to sign a model where we have
the hash computed from somewhere else and want to avoid the in-toto types.
"""

import sys

from typing_extensions import override

from model_signing.manifest import manifest as manifest_module
from model_signing.signing import signing


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class BytesPayload(signing.SigningPayload):
    """A payload that is a view into the bytes of a digest."""

    def __init__(self, digest: bytes):
        """Builds an instance of this payload.

        Don't call this directly in production. Use `from_manifest()` instead.

        Args:
            digest: The digest bytes as extracted from the manifest.
        """
        self.digest = digest

    @classmethod
    @override
    def from_manifest(cls, manifest: manifest_module.Manifest) -> Self:
        """Converts a manifest to the signing payload used for signing.

        The manifest must be a `DigestManifest` instance.

        Args:
            manifest: the manifest to convert to signing payload.

        Returns:
            An instance of `BytesPayload`.

        Raises:
            TypeError: If the manifest is not `DigestManifest`.
        """
        if not isinstance(manifest, manifest_module.DigestManifest):
            raise TypeError("Only DigestManifest is supported")

        # guaranteed to have exactly one item
        subject = list(manifest.resource_descriptors())[0]
        return cls(subject.digest.digest_value)



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Model serializers that operated at file level granularity."""

import abc
import base64
from collections.abc import Callable, Iterable
import concurrent.futures
import itertools
import pathlib
from typing import cast

from typing_extensions import override

from model_signing.hashing import file
from model_signing.hashing import hashing
from model_signing.manifest import manifest
from model_signing.serialization import serialization


def check_file_or_directory(
    path: pathlib.Path, *, allow_symlinks: bool = False
) -> None:
    """Checks that the given path is either a file or a directory.

    There is no support for sockets, pipes, or any other operating system
    concept abstracted as a file.

    Furthermore, this would raise if the path is a broken symlink, if it doesn't
    exists or if there are permission errors.

    Args:
        path: The path to check.
        allow_symlinks: Controls whether symbolic links are included. If a
          symlink is present but the flag is `False` (default) the
          serialization would raise an error.

    Raises:
        ValueError: The path is neither a file or a directory, or the path
          is a symlink and `allow_symlinks` is false.
    """
    if not allow_symlinks and path.is_symlink():
        raise ValueError(
            f"Cannot use '{path}' because it is a symlink. This"
            " behavior can be changed with `allow_symlinks`."
        )
    if not (path.is_file() or path.is_dir()):
        raise ValueError(
            f"Cannot use '{path}' as file or directory. It could be a"
            " special file, it could be missing, or there might be a"
            " permission issue."
        )


def _build_header(*, entry_name: str, entry_type: str) -> bytes:
    """Builds a header to encode a path with given name and type.

    Args:
        entry_name: The name of the entry to build the header for.
        entry_type: The type of the entry (file or directory).

    Returns:
        A sequence of bytes that encodes all arguments as a sequence of UTF-8
        bytes. Each argument is separated by dots and the last byte is also a
        dot (so the file digest can be appended unambiguously).
    """
    encoded_type = entry_type.encode("utf-8")
    # Prevent confusion if name has a "." inside by encoding to base64.
    encoded_name = base64.b64encode(entry_name.encode("utf-8"))
    # Note: empty string at the end, to terminate header with a "."
    return b".".join([encoded_type, encoded_name, b""])


def _ignored(path: pathlib.Path, ignore_paths: Iterable[pathlib.Path]) -> bool:
    """Determines if the provided path should be ignored.

    Args:
        path: The path to check.
        ignore_paths: The paths to ignore while serializing a model.

    Returns:
        Whether or not the provided path should be ignored.
    """
    return any(path.is_relative_to(ignore_path) for ignore_path in ignore_paths)


class FilesSerializer(serialization.Serializer):
    """Generic file serializer.

    Traverses the model directory and creates digests for every file found,
    possibly in parallel.

    Subclasses can then create a manifest with these digests, either listing
    them item by item, or combining everything into a single digest.
    """

    def __init__(
        self,
        file_hasher_factory: Callable[[pathlib.Path], file.FileHasher],
        *,
        max_workers: int | None = None,
        allow_symlinks: bool = False,
    ):
        """Initializes an instance to serialize a model with this serializer.

        Args:
            file_hasher_factory: A callable to build the hash engine used to
              hash individual files.
            max_workers: Maximum number of workers to use in parallel. Default
              is to defer to the `concurrent.futures` library.
            allow_symlinks: Controls whether symbolic links are included. If a
              symlink is present but the flag is `False` (default) the
              serialization would raise an error.
        """
        self._hasher_factory = file_hasher_factory
        self._max_workers = max_workers
        self._allow_symlinks = allow_symlinks

    @override
    def serialize(
        self,
        model_path: pathlib.Path,
        *,
        ignore_paths: Iterable[pathlib.Path] = frozenset(),
    ) -> manifest.Manifest:
        """Serializes the model given by the `model_path` argument.

        Args:
            model_path: The path to the model.
            ignore_paths: The paths to ignore during serialization. If a
              provided path is a directory, all children of the directory are
              ignored.

        Returns:
            The model's serialized `manifest.Manifest`

        Raises:
            ValueError: The model contains a symbolic link, but the serializer
              was not initialized with `allow_symlinks=True`.
        """
        paths = []
        # TODO: github.com/sigstore/model-transparency/issues/200 - When
        # Python3.12 is the minimum supported version, the glob can be replaced
        # with `pathlib.Path.walk` for a clearer interface, and some speed
        # improvement.
        for path in itertools.chain((model_path,), model_path.glob("**/*")):
            check_file_or_directory(path, allow_symlinks=self._allow_symlinks)
            if path.is_file() and not _ignored(path, ignore_paths):
                paths.append(path)

        manifest_items = []
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self._max_workers
        ) as tpe:
            futures = [
                tpe.submit(self._compute_hash, model_path, path)
                for path in paths
            ]
            for future in concurrent.futures.as_completed(futures):
                manifest_items.append(future.result())

        return self._build_manifest(manifest_items)

    def _compute_hash(
        self, model_path: pathlib.Path, path: pathlib.Path
    ) -> manifest.FileManifestItem:
        """Produces the manifest item of the file given by `path`.

        Args:
            model_path: The path to the model.
            path: Path to the file in the model, that is currently transformed
              to a manifest item.

        Returns:
            The itemized manifest.
        """
        relative_path = path.relative_to(model_path)
        digest = self._hasher_factory(path).compute()
        return manifest.FileManifestItem(path=relative_path, digest=digest)

    @abc.abstractmethod
    def _build_manifest(
        self, items: Iterable[manifest.FileManifestItem]
    ) -> manifest.Manifest:
        """Builds the manifest representing the serialization of the model."""
        pass


class ManifestSerializer(FilesSerializer):
    """Model serializer that produces an itemized manifest, at file level.

    Since the manifest lists each item individually, this will also enable
    support for incremental updates (to be added later).
    """

    @override
    def serialize(
        self,
        model_path: pathlib.Path,
        *,
        ignore_paths: Iterable[pathlib.Path] = frozenset(),
    ) -> manifest.FileLevelManifest:
        """Serializes the model given by the `model_path` argument.

        The only reason for the override is to change the return type, to be
        more restrictive. This is to signal that the only manifests that can be
        returned are `manifest.FileLevelManifest` instances.

        Args:
            model_path: The path to the model.
            ignore_paths: The paths to ignore during serialization. If a
              provided path is a directory, all children of the directory are
              ignored.

        Returns:
            The model's serialized `manifest.FileLevelManifest`

        Raises:
            ValueError: The model contains a symbolic link, but the serializer
              was not initialized with `allow_symlinks=True`.
        """
        return cast(
            manifest.FileLevelManifest,
            super().serialize(model_path, ignore_paths=ignore_paths),
        )

    @override
    def _build_manifest(
        self, items: Iterable[manifest.FileManifestItem]
    ) -> manifest.FileLevelManifest:
        return manifest.FileLevelManifest(items)


class _FileDigestTree:
    """A tree of files with their digests.

    Every leaf in the tree is a file, paired with its digest. Every intermediate
    node represents a directory. We need to pair every directory with a digest,
    in a bottom-up fashion.
    """

    def __init__(
        self, path: pathlib.PurePath, digest: hashing.Digest | None = None
    ):
        """Builds a node in the digest tree.

        Don't call this from outside of the class. Instead, use `build_tree`.

        Args:
            path: Path included in the node.
            digest: Optional hash of the path. Files must have a digest,
              directories never have one.
        """
        self._path = path
        self._digest = digest
        self._children: list[_FileDigestTree] = []

    @classmethod
    def build_tree(
        cls, items: Iterable[manifest.FileManifestItem]
    ) -> "_FileDigestTree":
        """Builds a tree out of the sequence of manifest items."""
        path_to_node: dict[pathlib.PurePath, _FileDigestTree] = {}

        for file_item in items:
            file = file_item.path
            node = cls(file, file_item.digest)
            for parent in file.parents:
                if parent in path_to_node:
                    parent_node = path_to_node[parent]
                    parent_node._children.append(node)
                    break  # everything else already exists

                parent_node = cls(parent)  # no digest for directories
                parent_node._children.append(node)
                path_to_node[parent] = parent_node
                node = parent_node

        # Handle empty model
        if not path_to_node:
            return cls(pathlib.PurePosixPath())

        return path_to_node[pathlib.PurePosixPath()]

    def get_digest(
        self, hasher_factory: Callable[[], hashing.StreamingHashEngine]
    ) -> hashing.Digest:
        """Returns the digest of this tree of files.

        Args:
            hasher_factory: A callable that returns a
              `hashing.StreamingHashEngine` instance used to merge individual
              digests to compute an aggregate digest.
        """
        hasher = hasher_factory()

        for child in sorted(self._children, key=lambda c: c._path):
            name = child._path.name
            if child._digest is not None:
                header = _build_header(entry_name=name, entry_type="file")
                hasher.update(header)
                hasher.update(child._digest.digest_value)
            else:
                header = _build_header(entry_name=name, entry_type="dir")
                hasher.update(header)
                digest = child.get_digest(hasher_factory)
                hasher.update(digest.digest_value)

        return hasher.compute()


class DigestSerializer(FilesSerializer):
    """Serializer for a model that performs a traversal of the model directory.

    This serializer produces a single hash for the entire model. If the model is
    a file, the hash is the digest of the file. If the model is a directory, we
    perform a depth-first traversal of the directory, hash each individual files
    and aggregate the hashes together.

    Currently, this has a different initialization than `FilesSerializer`, but
    this will likely change in a subsequent change. Similarly, currently, this
    only supports one single worker, but this will change in the future.
    """

    def __init__(
        self,
        file_hasher: file.SimpleFileHasher,
        merge_hasher_factory: Callable[[], hashing.StreamingHashEngine],
        *,
        allow_symlinks: bool = False,
    ):
        """Initializes an instance to serialize a model with this serializer.

        Args:
            hasher: The hash engine used to hash the individual files.
            merge_hasher_factory: A callable that returns a
              `hashing.StreamingHashEngine` instance used to merge individual
              file digests to compute an aggregate digest.
            allow_symlinks: Controls whether symbolic links are included. If a
              symlink is present but the flag is `False` (default) the
              serialization would raise an error.
        """

        def _factory(path: pathlib.Path) -> file.FileHasher:
            file_hasher.set_file(path)
            return file_hasher

        super().__init__(_factory, max_workers=1, allow_symlinks=allow_symlinks)
        self._merge_hasher_factory = merge_hasher_factory

    @override
    def serialize(
        self,
        model_path: pathlib.Path,
        *,
        ignore_paths: Iterable[pathlib.Path] = frozenset(),
    ) -> manifest.DigestManifest:
        """Serializes the model given by the `model_path` argument.

        The only reason for the override is to change the return type, to be
        more restrictive. This is to signal that the only manifests that can be
        returned are `manifest.DigestManifest` instances.

        Args:
            model_path: The path to the model.
            ignore_paths: The paths to ignore during serialization. If a
              provided path is a directory, all children of the directory are
              ignored.

        Returns:
            The model's serialized `manifest.DigestManifest`

        Raises:
            ValueError: The model contains a symbolic link, but the serializer
              was not initialized with `allow_symlinks=True`.
        """
        return cast(
            manifest.DigestManifest,
            super().serialize(model_path, ignore_paths=ignore_paths),
        )

    @override
    def _build_manifest(
        self, items: Iterable[manifest.FileManifestItem]
    ) -> manifest.DigestManifest:
        # Note: we do several computations here to try and match the old
        # behavior but these would be simplified in the future. Since we are
        # defining the hashing behavior, we can freely change this.

        # If the model is just one file, return the hash of the file.
        # A model is a file if we have one item only and its path is empty.
        items = list(items)
        if len(items) == 1 and not items[0].path.name:
            return manifest.DigestManifest(items[0].digest)

        # Otherwise, build a tree of files and compute the digests.
        tree = _FileDigestTree.build_tree(items)
        digest = tree.get_digest(self._merge_hasher_factory)
        return manifest.DigestManifest(digest)



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Signing payloads for models as in-toto statements.

To generate the signing payload we convert model manifests to in-toto formats,
as described by https://github.com/in-toto/attestation/tree/main/spec/v1. The
envelope format is DSSE, see https://github.com/secure-systems-lab/dsse.
"""

import pathlib
import sys
from typing import Any, Final

from in_toto_attestation.v1 import statement
from typing_extensions import override

from model_signing.hashing import hashing
from model_signing.hashing import memory
from model_signing.manifest import manifest as manifest_module
from model_signing.signing import signing


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class IntotoPayload(signing.SigningPayload):
    """A generic payload in in-toto format.

    This class is abstract for now as we will support multiple payload formats
    below.

    Each subclass defines a constant for the predicate type class attribute
    defined below.
    """

    predicate_type: Final[str]
    statement: Final[statement.Statement]

    @classmethod
    def manifest_from_payload(
        cls, payload: dict[str, Any]
    ) -> manifest_module.Manifest:
        """Builds a manifest from an in-memory in-toto payload.

        Delegates to all known subclasses until one matches the provided
        `predicateType` (matching `predicate_type` class attribute).

        Args:
            payload: the in memory in-toto payload to build a manifest from.

        Returns:
            A manifest that can be converted back to the same payload.

        Raises:
            ValueError: If the payload cannot be converted.
        """
        predicate_type = payload["predicateType"]
        subclasses = [
            SingleDigestIntotoPayload,
            DigestOfDigestsIntotoPayload,
            DigestOfShardDigestsIntotoPayload,
            DigestsIntotoPayload,
            ShardDigestsIntotoPayload,
        ]

        for subcls in subclasses:
            if predicate_type == subcls.predicate_type:
                return subcls.manifest_from_payload(payload)

        raise ValueError(f"Unknown in-toto predicate type {predicate_type}")


class SingleDigestIntotoPayload(IntotoPayload):
    """In-toto payload where the model is serialized to just one digest.

    In this case, we encode the model as the only subject of the statement. We
    set the name field to ".", and use the digest as the one resulting from the
    model serialization.

    However, since we use custom hashing algorithms, but these are not supported
    by existing tools, we claim that the digest algorithm is sha-256 and include
    the real digest in the predicate.

    Example:
    ```json
    {
      "_type": "https://in-toto.io/Statement/v1",
      "subject": [
        {
          "name": ".",
          "digest": {
            "sha256": "3aab065c...."
          }
        }
      ],
      "predicateType": "https://model_signing/Digest/v0.1",
      "predicate": {
        "actual_hash_algorithm": "file-sha256"
      }
    }
    ```

    If the predicate is missing (or does not set "actual_hash_algorithm"), it
    should be assumed that the digest is actually computed via the algorithm
    present in the resource descriptor (i.e., sha256).

    See also https://github.com/sigstore/sigstore-python/issues/1018.
    """

    predicate_type: Final[str] = "https://model_signing/Digest/v0.1"

    def __init__(self, *, digest_hex: str, digest_algorithm: str):
        """Builds an instance of this in-toto payload.

        Don't call this directly in production. Use `from_manifest()` instead.

        Args:
            digest_hex: the hexadecimal, human readable, digest of the subject.
            digest_algorithm: the algorithm used to compute the digest.
        """
        digest = {"sha256": digest_hex}
        descriptor = statement.ResourceDescriptor(name=".", digest=digest).pb

        self.statement = statement.Statement(
            subjects=[descriptor],
            predicate_type=self.predicate_type,
            predicate={"actual_hash_algorithm": digest_algorithm},
        )

    @classmethod
    @override
    def from_manifest(cls, manifest: manifest_module.Manifest) -> Self:
        """Converts a manifest to the signing payload used for signing.

        The manifest must be a `DigestManifest` instance.

        Args:
            manifest: the manifest to convert to signing payload.

        Returns:
            An instance of `SingleDigestIntotoPayload`.

        Raises:
            TypeError: If the manifest is not `DigestManifest`.
        """
        if not isinstance(manifest, manifest_module.DigestManifest):
            raise TypeError("Only DigestManifest is supported")

        # guaranteed to have exactly one item
        subject = list(manifest.resource_descriptors())[0]
        digest = subject.digest
        return cls(
            digest_hex=digest.digest_hex, digest_algorithm=digest.algorithm
        )

    @classmethod
    @override
    def manifest_from_payload(
        cls, payload: dict[str, Any]
    ) -> manifest_module.DigestManifest:
        """Builds a manifest from an in-memory in-toto payload.

        Args:
            payload: the in memory in-toto payload to build a manifest from.

        Returns:
            A manifest that can be converted back to the same payload.

        Raises:
            ValueError: If the payload does not match the expected payload
              format for this class. See `from_manifest`.
        """
        subjects = payload["subject"]
        predicate = payload["predicate"]

        if len(subjects) != 1:
            raise ValueError(f"Expected one single subject, got {subjects}")

        algorithm = predicate["actual_hash_algorithm"]
        digest_value = bytes.fromhex(subjects[0]["digest"]["sha256"])
        digest = hashing.Digest(algorithm, digest_value)
        return manifest_module.DigestManifest(digest)


def _convert_descriptors_to_hashed_statement(
    manifest: manifest_module.Manifest,
    *,
    predicate_type: str,
    predicate_top_level_name: str,
) -> statement.Statement:
    """Converts manifest descriptors to an in-toto statement with payload.

    Args:
        manifest: The manifest to extract the descriptors from. Assumed valid.
        predicate_type: The predicate_type to use in the in-toto statement.
        predicate_top_level_name: Name to use in the payload for the array of
          the subjects.
    """
    hasher = memory.SHA256()
    subjects = []
    for descriptor in manifest.resource_descriptors():
        hasher.update(descriptor.digest.digest_value)
        subjects.append(
            {
                "name": descriptor.identifier,
                "digest": descriptor.digest.digest_hex,
                "algorithm": descriptor.digest.algorithm,
            }
        )

    digest = {"sha256": hasher.compute().digest_hex}
    descriptor = statement.ResourceDescriptor(name=".", digest=digest).pb

    return statement.Statement(
        subjects=[descriptor],
        predicate_type=predicate_type,
        predicate={predicate_top_level_name: subjects},
    )


class DigestOfDigestsIntotoPayload(IntotoPayload):
    """In-toto payload where the subject is a digest of digests of model files.

    This payload is supposed to be used for manifests where every file in the
    model is matched with a digest. Because existing tooling only supports
    established hashing algorithms, we record every such digest in the predicate
    part and compute a hash for the subject by using sha256 on the concatenation
    of the file hashes. To ensure determinism, the hashes are sorted
    alphabetically by filename.

    Example:
    ```json
    {
      "_type": "https://in-toto.io/Statement/v1",
      "subject": [
        {
          "name": ".",
          "digest": {
            "sha256": "18b5a4..."
          }
        }
      ],
      "predicateType": "https://model_signing/DigestOfDigests/v0.1",
      "predicate": {
        "files": [
          {
            "digest": "6efa14...",
            "algorithm": "file-sha256",
            "name": "d0/d1/d2/d3/d4/f0"
          },
          {
            "digest": "a9bc14...",
            "algorithm": "file-sha256",
            "name": "d0/d1/d2/d3/d4/f1"
          },
          {
            "digest": "5f597e...",
            "algorithm": "file-sha256",
            "name": "d0/d1/d2/d3/d4/f2"
          },
          {
            "digest": "eaf677...",
            "algorithm": "file-sha256",
            "name": "d0/d1/d2/d3/d4/f3"
          }
        ]
      }
    }
    ```

    A missing predicate, or a predicate for which an entry does not have valid
    name, digest, or algorithm should be considered invalid and fail integrity
    verification.

    See also https://github.com/sigstore/sigstore-python/issues/1018.
    """

    predicate_type: Final[str] = "https://model_signing/DigestOfDigests/v0.1"

    def __init__(self, statement: statement.Statement):
        """Builds an instance of this in-toto payload.

        Don't call this directly in production. Use `from_manifest()` instead.

        Args:
            statement: The DSSE statement representing this in-toto payload.
        """
        self.statement = statement

    @classmethod
    @override
    def from_manifest(cls, manifest: manifest_module.Manifest) -> Self:
        """Converts a manifest to the signing payload used for signing.

        The manifest must be one where every model file is paired with its own
        digest. Currently, this is only `FileLevelManifest`.

        Args:
            manifest: the manifest to convert to signing payload.

        Returns:
            An instance of `DigestOfDigestsIntotoPayload`.

        Raises:
            TypeError: If the manifest is not `FileLevelManifest`.
        """
        if not isinstance(manifest, manifest_module.FileLevelManifest):
            raise TypeError("Only FileLevelManifest is supported")

        statement = _convert_descriptors_to_hashed_statement(
            manifest,
            predicate_type=cls.predicate_type,
            predicate_top_level_name="files",
        )
        return cls(statement)

    @classmethod
    @override
    def manifest_from_payload(
        cls, payload: dict[str, Any]
    ) -> manifest_module.FileLevelManifest:
        """Builds a manifest from an in-memory in-toto payload.

        Args:
            payload: the in memory in-toto payload to build a manifest from.

        Returns:
            A manifest that can be converted back to the same payload.

        Raises:
            ValueError: If the payload does not match the expected payload
              format for this class. See `from_manifest`.
        """
        subjects = payload["subject"]
        predicate = payload["predicate"]

        if len(subjects) != 1:
            raise ValueError(f"Expected one single subject, got {subjects}")

        hasher = memory.SHA256()
        items = []
        for file in predicate["files"]:
            path = pathlib.PurePosixPath(file["name"])
            digest = hashing.Digest(
                file["algorithm"], bytes.fromhex(file["digest"])
            )
            item = manifest_module.FileManifestItem(path=path, digest=digest)
            items.append(item)
            hasher.update(digest.digest_value)

        expected_digest = subjects[0]["digest"]["sha256"]
        obtained_digest = hasher.compute().digest_hex
        if obtained_digest != expected_digest:
            raise ValueError(
                "Verification failed. "
                f"Expected {expected_digest}, got {obtained_digest}"
            )

        return manifest_module.FileLevelManifest(items)


class DigestOfShardDigestsIntotoPayload(IntotoPayload):
    """In-toto payload where the subject is a digest of digests of file shards.

    This payload is supposed to be used for manifests where every file shard in
    the model is matched with a digest. Because existing tooling only supports
    established hashing algorithms, we record every such digest in the predicate
    part and compute a hash for the subject by using sha256 on the concatenation
    of the shard hashes. To ensure determinism, the hashes are sorted
    by file shard (alphabetically by name, then ordered by start offset).

    Example:
    ```json
    {
      "_type": "https://in-toto.io/Statement/v1",
      "subject": [
        {
          "name": ".",
          "digest": {
            "sha256": "18b5a4..."
          }
        }
      ],
      "predicateType": "https://model_signing/DigestOfShardDigests/v0.1",
      "predicate": {
        "shards": [
          {
            "digest": "6efa14...",
            "algorithm": "file-sha256-1000000",
            "name": "d0/d1/d2/d3/d4/f0:0:16"
          },
          {
            "digest": "a9bc14...",
            "algorithm": "file-sha256-1000000",
            "name": "d0/d1/d2/d3/d4/f1:0:16"
          },
          {
            "digest": "5f597e...",
            "algorithm": "file-sha256-1000000",
            "name": "d0/d1/d2/d3/d4/f2:0:16"
          },
          {
            "digest": "eaf677...",
            "algorithm": "file-sha256-1000000",
            "name": "d0/d1/d2/d3/d4/f3:0:16"
          }
        ]
      }
    }
    ```

    A missing predicate, or a predicate for which an entry does not have valid
    name, digest, or algorithm should be considered invalid and fail integrity
    verification.

    See also https://github.com/sigstore/sigstore-python/issues/1018.
    """

    predicate_type: Final[str] = (
        "https://model_signing/DigestOfShardDigests/v0.1"
    )

    def __init__(self, statement: statement.Statement):
        """Builds an instance of this in-toto payload.

        Don't call this directly in production. Use `from_manifest()` instead.

        Args:
            statement: The DSSE statement representing this in-toto payload.
        """
        self.statement = statement

    @classmethod
    @override
    def from_manifest(cls, manifest: manifest_module.Manifest) -> Self:
        """Converts a manifest to the signing payload used for signing.

        The manifest must be one where every model shard is paired with its own
        digest. Currently, this is only `ShardLevelManifest`.

        Args:
            manifest: the manifest to convert to signing payload.

        Returns:
            An instance of `DigestOfDigestsIntotoPayload`.

        Raises:
            TypeError: If the manifest is not `ShardLevelManifest`.
        """
        if not isinstance(manifest, manifest_module.ShardLevelManifest):
            raise TypeError("Only ShardLevelManifest is supported")

        statement = _convert_descriptors_to_hashed_statement(
            manifest,
            predicate_type=cls.predicate_type,
            predicate_top_level_name="shards",
        )
        return cls(statement)

    @classmethod
    @override
    def manifest_from_payload(
        cls, payload: dict[str, Any]
    ) -> manifest_module.ShardLevelManifest:
        """Builds a manifest from an in-memory in-toto payload.

        Args:
            payload: the in memory in-toto payload to build a manifest from.

        Returns:
            A manifest that can be converted back to the same payload.

        Raises:
            ValueError: If the payload does not match the expected payload
              format for this class. See `from_manifest`.
        """
        subjects = payload["subject"]
        predicate = payload["predicate"]

        if len(subjects) != 1:
            raise ValueError(f"Expected one single subject, got {subjects}")

        hasher = memory.SHA256()
        items = []
        for entry in predicate["shards"]:
            shard = manifest_module.Shard.from_str(entry["name"])
            digest = hashing.Digest(
                entry["algorithm"], bytes.fromhex(entry["digest"])
            )
            item = manifest_module.ShardedFileManifestItem(
                path=shard.path, start=shard.start, end=shard.end, digest=digest
            )
            items.append(item)
            hasher.update(digest.digest_value)

        expected_digest = subjects[0]["digest"]["sha256"]
        obtained_digest = hasher.compute().digest_hex
        if obtained_digest != expected_digest:
            raise ValueError(
                "Verification failed. "
                f"Expected {expected_digest}, got {obtained_digest}"
            )

        return manifest_module.ShardLevelManifest(items)


def _convert_descriptors_to_direct_statement(
    manifest: manifest_module.Manifest, predicate_type: str
) -> statement.Statement:
    """Converts manifest descriptors to an in-toto statement, as subjects.

    Args:
        manifest: The manifest to extract the descriptors from. Assumed valid.
        predicate_type: The predicate_type to use in the in-toto statement.
    """
    subjects = []
    for descriptor in manifest.resource_descriptors():
        subject = statement.ResourceDescriptor(
            name=descriptor.identifier,
            digest={"sha256": descriptor.digest.digest_hex},
            annotations={"actual_hash_algorithm": descriptor.digest.algorithm},
        )
        subjects.append(subject.pb)

    return statement.Statement(
        subjects=subjects,
        predicate_type=predicate_type,
        # https://github.com/in-toto/attestation/issues/374
        predicate={"unused": "Unused, just passed due to API requirements"},
    )


class DigestsIntotoPayload(IntotoPayload):
    """In-toto payload where the subjects are the model files themselves.

    This payload is supposed to be used for manifests where every file in the
    model is matched with a digest. Because existing tooling only supports
    established hashing algorithms, we annotate every subject with the actual
    hash algorithm used to compute the file digest, and use "sha256" as the
    algorithm name in the digest itself.

    Example:
    ```json
    {
      "_type": "https://in-toto.io/Statement/v1",
      "subject": [
        {
          "name": "d0/d1/d2/d3/d4/f0",
          "digest": {
            "sha256": "6efa14..."
          },
          "annotations": {
            "actual_hash_algorithm": "file-sha256"
          }
        },
        {
          "name": "d0/d1/d2/d3/d4/f1",
          "digest": {
            "sha256": "a9bc14..."
          },
          "annotations": {
            "actual_hash_algorithm": "file-sha256"
          }
        },
        {
          "name": "d0/d1/d2/d3/d4/f2",
          "digest": {
            "sha256": "5f597e..."
          },
          "annotations": {
            "actual_hash_algorithm": "file-sha256"
          }
        },
        {
          "name": "d0/d1/d2/d3/d4/f3",
          "digest": {
            "sha256": "eaf677..."
          },
          "annotations": {
            "actual_hash_algorithm": "file-sha256"
          }
        }
      ],
      "predicateType": "https://model_signing/Digests/v0.1",
      "predicate": {
        "unused": "Unused, just passed due to API requirements"
      }
    }
    ```

    If the annotation for a subject is missing, or it does not contain
    actual_hash_algorithm, it should be assumed that the digest is computed via
    the algorithm listed in the digest dictionary (i.e., sha256).

    See also https://github.com/sigstore/sigstore-python/issues/1018.
    """

    predicate_type: Final[str] = "https://model_signing/Digests/v0.1"

    def __init__(self, statement: statement.Statement):
        """Builds an instance of this in-toto payload.

        Don't call this directly in production. Use `from_manifest()` instead.

        Args:
            statement: The DSSE statement representing this in-toto payload.
        """
        self.statement = statement

    @classmethod
    @override
    def from_manifest(cls, manifest: manifest_module.Manifest) -> Self:
        """Converts a manifest to the signing payload used for signing.

        The manifest must be one where every model file is paired with its own
        digest. Currently, this is only `FileLevelManifest`.

        Args:
            manifest: the manifest to convert to signing payload.

        Returns:
            An instance of `DigestOfDigestsIntotoPayload`.

        Raises:
            TypeError: If the manifest is not `FileLevelManifest`.
        """
        if not isinstance(manifest, manifest_module.FileLevelManifest):
            raise TypeError("Only FileLevelManifest is supported")

        statement = _convert_descriptors_to_direct_statement(
            manifest, predicate_type=cls.predicate_type
        )
        return cls(statement)

    @classmethod
    @override
    def manifest_from_payload(
        cls, payload: dict[str, Any]
    ) -> manifest_module.FileLevelManifest:
        """Builds a manifest from an in-memory in-toto payload.

        Args:
            payload: the in memory in-toto payload to build a manifest from.

        Returns:
            A manifest that can be converted back to the same payload.

        Raises:
            ValueError: If the payload does not match the expected payload
              format for this class. See `from_manifest`.
        """
        subjects = payload["subject"]

        items = []
        for subject in subjects:
            path = pathlib.PurePosixPath(subject["name"])
            algorithm = subject["annotations"]["actual_hash_algorithm"]
            digest_value = subject["digest"]["sha256"]
            digest = hashing.Digest(algorithm, bytes.fromhex(digest_value))
            item = manifest_module.FileManifestItem(path=path, digest=digest)
            items.append(item)

        return manifest_module.FileLevelManifest(items)


class ShardDigestsIntotoPayload(IntotoPayload):
    """In-toto payload where the subjects are the model shards themselves.

    This payload is supposed to be used for manifests where every file shard in
    the model is matched with a digest. Because existing tooling only supports
    established hashing algorithms, we annotate every subject with the actual
    hash algorithm used to compute the file digest, and use "sha256" as the
    algorithm name in the digest itself.

    Example:
    ```json
    {
      "_type": "https://in-toto.io/Statement/v1",
      "subject": [
        {
          "name": "d0/d1/d2/d3/d4/f0:0:16",
          "digest": {
            "sha256": "6efa14..."
          },
          "annotations": {
            "actual_hash_algorithm": "file-sha256-1000000"
          }
        },
        {
          "name": "d0/d1/d2/d3/d4/f1:0:16",
          "digest": {
            "sha256": "a9bc14..."
          },
          "annotations": {
            "actual_hash_algorithm": "file-sha256-1000000"
          }
        },
        {
          "name": "d0/d1/d2/d3/d4/f2:0:16",
          "digest": {
            "sha256": "5f597e..."
          },
          "annotations": {
            "actual_hash_algorithm": "file-sha256-1000000"
          }
        },
        {
          "name": "d0/d1/d2/d3/d4/f3:0:16",
          "digest": {
            "sha256": "eaf677..."
          },
          "annotations": {
            "actual_hash_algorithm": "file-sha256-1000000"
          }
        }
      ],
      "predicateType": "https://model_signing/ShardDigests/v0.1",
      "predicate": {
        "unused": "Unused, just passed due to API requirements"
      }
    }
    ```

    If the annotation for a subject is missing, or it does not contain
    actual_hash_algorithm, it should be assumed that the digest is computed via
    the algorithm listed in the digest dictionary (i.e., sha256).

    See also https://github.com/sigstore/sigstore-python/issues/1018.
    """

    predicate_type: Final[str] = "https://model_signing/ShardDigests/v0.1"

    def __init__(self, statement: statement.Statement):
        """Builds an instance of this in-toto payload.

        Don't call this directly in production. Use `from_manifest()` instead.

        Args:
            statement: The DSSE statement representing this in-toto payload.
        """
        self.statement = statement

    @classmethod
    @override
    def from_manifest(cls, manifest: manifest_module.Manifest) -> Self:
        """Converts a manifest to the signing payload used for signing.

        The manifest must be one where every model shard is paired with its own
        digest. Currently, this is only `ShardLevelManifest`.

        Args:
            manifest: the manifest to convert to signing payload.

        Returns:
            An instance of `DigestOfDigestsIntotoPayload`.

        Raises:
            TypeError: If the manifest is not `ShardLevelManifest`.
        """
        if not isinstance(manifest, manifest_module.ShardLevelManifest):
            raise TypeError("Only ShardLevelManifest is supported")

        statement = _convert_descriptors_to_direct_statement(
            manifest, predicate_type=cls.predicate_type
        )
        return cls(statement)

    @classmethod
    @override
    def manifest_from_payload(
        cls, payload: dict[str, Any]
    ) -> manifest_module.ShardLevelManifest:
        """Builds a manifest from an in-memory in-toto payload.

        Args:
            payload: the in memory in-toto payload to build a manifest from.

        Returns:
            A manifest that can be converted back to the same payload.

        Raises:
            ValueError: If the payload does not match the expected payload
              format for this class. See `from_manifest`.
        """
        subjects = payload["subject"]

        items = []
        for subject in subjects:
            shard = manifest_module.Shard.from_str(subject["name"])
            algorithm = subject["annotations"]["actual_hash_algorithm"]
            digest_value = subject["digest"]["sha256"]
            digest = hashing.Digest(algorithm, bytes.fromhex(digest_value))
            item = manifest_module.ShardedFileManifestItem(
                path=shard.path, start=shard.start, end=shard.end, digest=digest
            )
            items.append(item)

        return manifest_module.ShardLevelManifest(items)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functionality to sign and verify models with keys."""

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.hashes import SHA256
from google.protobuf import json_format
from in_toto_attestation.v1 import statement
from in_toto_attestation.v1 import statement_pb2 as statement_pb
from sigstore_protobuf_specs.dev.sigstore.bundle import v1 as bundle_pb
from sigstore_protobuf_specs.dev.sigstore.common import v1 as common_pb
from sigstore_protobuf_specs.io import intoto as intoto_pb

from model_signing.signature import encoding
from model_signing.signature.signing import Signer
from model_signing.signature.verifying import VerificationError
from model_signing.signature.verifying import Verifier


def load_ec_private_key(
    path: str, password: str | None = None
) -> ec.EllipticCurvePrivateKey:
    private_key: ec.EllipticCurvePrivateKey
    with open(path, "rb") as fd:
        serialized_key = fd.read()
    private_key = serialization.load_pem_private_key(
        serialized_key, password=password
    )
    return private_key


class ECKeySigner(Signer):
    """Provides a Signer using an elliptic curve private key for signing."""

    def __init__(self, private_key: ec.EllipticCurvePrivateKey):
        self._private_key = private_key

    @classmethod
    def from_path(cls, private_key_path: str, password: str | None = None):
        private_key = load_ec_private_key(private_key_path, password)
        return cls(private_key)

    def sign(self, stmnt: statement.Statement) -> bundle_pb.Bundle:
        pae = encoding.pae(stmnt.pb)
        sig = self._private_key.sign(pae, ec.ECDSA(SHA256()))
        env = intoto_pb.Envelope(
            payload=json_format.MessageToJson(stmnt.pb).encode(),
            payload_type=encoding.PAYLOAD_TYPE,
            signatures=[intoto_pb.Signature(sig=sig, keyid=None)],
        )
        bdl = bundle_pb.Bundle(
            media_type="application/vnd.dev.sigstore.bundle.v0.3+json",
            verification_material=bundle_pb.VerificationMaterial(
                public_key=common_pb.PublicKey(
                    raw_bytes=self._private_key.public_key().public_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PublicFormat.SubjectPublicKeyInfo,
                    ),
                    key_details=common_pb.PublicKeyDetails.PKIX_ECDSA_P256_SHA_256,
                )
            ),
            dsse_envelope=env,
        )

        return bdl


class ECKeyVerifier(Verifier):
    """Provides a verifier using a public key."""

    def __init__(self, public_key: ec.EllipticCurvePublicKey):
        self._public_key = public_key

    @classmethod
    def from_path(cls, key_path: str):
        with open(key_path, "rb") as fd:
            serialized_key = fd.read()
        public_key = serialization.load_pem_public_key(serialized_key)
        return cls(public_key)

    def verify(self, bundle: bundle_pb.Bundle) -> None:
        statement = json_format.Parse(
            bundle.dsse_envelope.payload,
            statement_pb.Statement(),  # pylint: disable=no-member
        )
        pae = encoding.pae(statement)
        try:
            self._public_key.verify(
                bundle.dsse_envelope.signatures[0].sig, pae, ec.ECDSA(SHA256())
            )
        except Exception as e:
            raise VerificationError(
                "signature verification failed " + str(e)
            ) from e



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Machinery for computing digests for a single object.

We define an abstract `HashEngine` class which can be used in type annotations
and is at the root of the hashing classes hierarchy.

Since there are multiple hashing methods that we support, users should always
specify the algorithm and the digest value.
"""

import abc
import dataclasses
from typing import Protocol


@dataclasses.dataclass(frozen=True)
class Digest:
    """A digest computed by a `HashEngine`."""

    algorithm: str
    digest_value: bytes

    @property
    def digest_hex(self) -> str:
        """Hexadecimal, human readable, equivalent of `digest`."""
        return self.digest_value.hex()

    @property
    def digest_size(self) -> int:
        """The size, in bytes, of the digest."""
        return len(self.digest_value)


class HashEngine(metaclass=abc.ABCMeta):
    """Generic hash engine."""

    @abc.abstractmethod
    def compute(self) -> Digest:
        """Computes the digest of data passed to the engine."""
        pass

    @property
    @abc.abstractmethod
    def digest_name(self) -> str:
        """The canonical name of the algorithm used to compute the hash.

        Subclasses MUST use the `digest_name()` method to record all parameters
        that influence the hash output. For example, if a file is split into
        shards which are hashed separately and the final digest value is
        computed by aggregating these hashes, then the shard size must be given
        in the output string.
        """
        pass

    @property
    @abc.abstractmethod
    def digest_size(self) -> int:
        """The size, in bytes, of the digests produced by the engine."""
        pass


class Streaming(Protocol):
    """A protocol to support streaming data to `HashEngine` objects."""

    @abc.abstractmethod
    def update(self, data: bytes) -> None:
        """Appends additional bytes to the data to be hashed."""
        pass

    @abc.abstractmethod
    def reset(self, data: bytes = b"") -> None:
        """Resets the data to be hashed to the passed argument."""
        pass


class StreamingHashEngine(Streaming, HashEngine):
    """A `HashEngine` that can stream data to be hashed."""

    pass



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Empty signing infrastructure.

This is only used to test the signing and verification machinery. It can also be
used as a default implementation in cases where some of the machinery doesn't
need to do anything (e.g., in testing or in cases where verification is being
done from outside the library).
"""

import pathlib
import sys

from typing_extensions import override

from model_signing.manifest import manifest
from model_signing.signing import signing


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class EmptySigningPayload(signing.SigningPayload):
    """An empty signing payload, mostly just for testing."""

    @classmethod
    @override
    def from_manifest(cls, manifest: manifest.Manifest) -> Self:
        """Converts a manifest to the signing payload used for signing.

        Args:
            manifest: the manifest to convert to signing payload.

        Returns:
            An instance of `EmptySigningPayload`.
        """
        del manifest  # unused
        return cls()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented

        return True  # all instances are equal


class EmptySignature(signing.Signature):
    """Empty signature, mostly for testing.

    Can also be used in cases where the signing result does not need to
    follow the rest of the signing machinery in this library (e.g., it is
    verified only by tooling that assume a different flow, or the existing
    signing machinery already manages writing signatures as a side effect of the
    signing process).
    """

    @override
    def write(self, path: pathlib.Path) -> None:
        """Writes the signature to disk, to the given path.

        Since the signature is empty this function actually does nothing, it's
        here just to match the API.

        Args:
            path: the path to write the signature to. Ignored.
        """
        del path  # unused

    @classmethod
    @override
    def read(cls, path: pathlib.Path) -> Self:
        """Reads the signature from disk.

        Since the signature is empty, this does nothing besides just returning
        an instance of `EmptySignature`.

        Args:
            path: the path to read the signature from. Ignored.

        Returns:
            An instance of `EmptySignature`.
        """
        del path  # unused
        return cls()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented

        return True  # all instances are equal


class EmptySigner(signing.Signer):
    """A signer that only produces `EmptySignature` objects, for testing."""

    @override
    def sign(self, payload: signing.SigningPayload) -> EmptySignature:
        """Signs the provided signing payload.

        Args:
            payload: the `SigningPayload` instance that should be signed.

        Returns:
            An `EmptySignature` object.
        """
        del payload  # unused
        return EmptySignature()


class EmptyVerifier(signing.Verifier):
    """Verifier that accepts only `EmptySignature` objects.

    Rather than producing a manifest out of thin air, the verifier also fails to
    verify the signature, even if it is in the accepted `EmptySignature` format.
    """

    @override
    def verify(self, signature: signing.Signature) -> manifest.Manifest:
        """Verifies the signature.

        Args:
            signature: the signature to verify.

        Raises:
            TypeError: If the signature is not an `EmptySignature` instance.
            ValueError: If the signature is an `EmptySignature` instance. This
              simulates failing signature verification.
        """
        if isinstance(signature, EmptySignature):
            raise ValueError("Signature verification failed")
        raise TypeError("Only `EmptySignature` instances are supported")



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This package provides the functionality to verify signed models."""

import abc

from sigstore_protobuf_specs.dev.sigstore.bundle import v1 as bundle_pb


class VerificationError(Exception):
    """Typed verification error to provide error handling information."""

    def __init__(self, message: str) -> None:
        super().__init__(message)


class Verifier(abc.ABC):
    """Verifier is the abstract base class for all verifying methods."""

    @abc.abstractmethod
    def verify(self, bundle: bundle_pb.Bundle) -> None:
        """Verify the signature of the provided bundle.

        Args:
            bundle (bundle_pb.Bundle): the bundle that needs to be verified.

        Raises:
            VerificationError: verification failure exception.
        """
